# -*- coding: utf-8 -*-

import os
import json
import logging

import sqlalchemy
from sqlalchemy import func, exc, and_, or_, not_
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, BigInteger, String, Text, Boolean, Date, DateTime, Numeric
from sqlalchemy.orm.exc import NoResultFound

from litepipeline.manager.utils import common
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger(__name__)


BaseApplications = declarative_base()

class ApplicationsTable(BaseApplications):
    __tablename__ = "applications"

    id = Column(Integer, primary_key = True, autoincrement = True)
    application_id = Column(Text, unique = True, nullable = False, index = True)
    name = Column(Text, nullable = False, index = True)
    create_at = Column(DateTime, nullable = False, index = True)
    update_at = Column(DateTime, nullable = False, index = True)
    sha1 = Column(Text, nullable = False)
    description = Column(Text)

    @classmethod
    def init_engine_and_session(cls):
        cls.engine = create_engine('sqlite:///' + os.path.join(CONFIG["data_path"], "applications.db"), echo = False)
        cls.session = sessionmaker(bind = cls.engine)
        return cls.engine, cls.session

    def to_dict(self):
        return {
            "id": self.id,
            "application_id": self.application_id,
            "name": self.name,
            "create_at": str(self.create_at), # "%Y-%m-%d %H:%M:%S.%f")
            "update_at": str(self.update_at),
            "sha1": self.sha1,
            "description": self.description,
        }

    def parse_dict(self, source):
        result = False

        attrs = [
            "application_id",
            "name",
            "create_at",
            "update_at",
            "sha1",
            "description",
        ]

        if hasattr(source, "__getitem__"):
            for attr in attrs:
                try:
                    setattr(self, attr, source[attr])
                except:
                    LOG.debug("some exception occured when extract %s attribute to object, i will discard it",
                        attr)
                    continue
            result = True
        else:
            LOG.debug("input param source does not have dict-like method, so i will do nothing at all!")
        return result

    def __repr__(self):
        return "application: %s" % self.to_dict()


BaseApplicationHistory = declarative_base()

class ApplicationHistoryTable(BaseApplicationHistory):
    __tablename__ = "application_history"

    id = Column(Integer, primary_key = True, autoincrement = True)
    application_id = Column(Text, nullable = False, index = True)
    create_at = Column(DateTime, nullable = False, index = True)
    sha1 = Column(Text, nullable = False, index = True)
    description = Column(Text)

    @classmethod
    def init_engine_and_session(cls):
        cls.engine = create_engine('sqlite:///' + os.path.join(CONFIG["data_path"], "application_history.db"), echo = False)
        cls.session = sessionmaker(bind = cls.engine)
        return cls.engine, cls.session

    def to_dict(self):
        return {
            "id": self.id,
            "application_id": self.application_id,
            "create_at": str(self.create_at), # "%Y-%m-%d %H:%M:%S.%f")
            "sha1": self.sha1,
            "description": self.description,
        }

    def parse_dict(self, source):
        result = False

        attrs = [
            "application_id",
            "create_at",
            "sha1",
            "description",
        ]

        if hasattr(source, "__getitem__"):
            for attr in attrs:
                try:
                    setattr(self, attr, source[attr])
                except:
                    LOG.debug("some exception occured when extract %s attribute to object, i will discard it",
                        attr)
                    continue
            result = True
        else:
            LOG.debug("input param source does not have dict-like method, so i will do nothing at all!")
        return result

    def __repr__(self):
        return "application_history: %s" % self.to_dict()


BaseTasks = declarative_base()

class TasksTable(BaseTasks):
    __tablename__ = "tasks"

    id = Column(Integer, primary_key = True, autoincrement = True)
    task_id = Column(Text, unique = True, nullable = False, index = True)
    task_name = Column(Text, nullable = False, index = True)
    application_id = Column(Text, nullable = False, index = True)
    work_id = Column(Text, nullable = False, index = True)
    service_id = Column(Text, nullable = False, index = True)
    create_at = Column(DateTime, nullable = False, index = True)
    start_at = Column(DateTime, index = True)
    update_at = Column(DateTime, nullable = False, index = True)
    end_at = Column(DateTime, index = True)
    stage = Column(Text, nullable = False, index = True)
    status = Column(Text, index = True)
    input_data = Column(Text)
    result = Column(Text)

    @classmethod
    def init_engine_and_session(cls):
        cls.engine = create_engine('sqlite:///' + os.path.join(CONFIG["data_path"], "tasks.db"), echo = False)
        cls.session = sessionmaker(bind = cls.engine)
        return cls.engine, cls.session

    def to_dict(self):
        return {
            "id": self.id,
            "task_id": self.task_id,
            "task_name": self.task_name,
            "application_id": self.application_id,
            "work_id": self.work_id,
            "service_id": self.service_id,
            "create_at": str(self.create_at),
            "start_at": str(self.start_at) if self.start_at else self.start_at,
            "update_at": str(self.update_at),
            "end_at": str(self.end_at) if self.end_at else self.end_at,
            "stage": self.stage,
            "status": self.status,
            "input_data": json.loads(self.input_data),
            "result": json.loads(self.result),
        }

    def parse_dict(self, source):
        result = False

        attrs = [
            "task_id",
            "task_name",
            "application_id",
            "work_id",
            "service_id",
            "create_at",
            "start_at",
            "update_at",
            "end_at",
            "stage",
            "status",
            "input_data",
            "result",
        ]

        if hasattr(source, "__getitem__"):
            for attr in attrs:
                try:
                    setattr(self, attr, source[attr])
                except:
                    LOG.debug("some exception occured when extract %s attribute to object, i will discard it",
                        attr)
                    continue
            result = True
        else:
            LOG.debug("input param source does not have dict-like method, so i will do nothing at all!")
        return result

    def __repr__(self):
        return "task: %s" % self.to_dict()


BaseWorkflows = declarative_base()

class WorkflowsTable(BaseWorkflows):
    __tablename__ = "workflows"

    id = Column(Integer, primary_key = True, autoincrement = True)
    workflow_id = Column(Text, unique = True, nullable = False, index = True)
    name = Column(Text, nullable = False, index = True)
    create_at = Column(DateTime, nullable = False, index = True)
    update_at = Column(DateTime, nullable = False, index = True)
    configuration = Column(Text, nullable = False)
    description = Column(Text)
    enable = Column(Boolean, nullable = False)

    @classmethod
    def init_engine_and_session(cls):
        cls.engine = create_engine('sqlite:///' + os.path.join(CONFIG["data_path"], "workflows.db"), echo = False)
        cls.session = sessionmaker(bind = cls.engine)
        return cls.engine, cls.session

    def to_dict(self):
        return {
            "id": self.id,
            "workflow_id": self.workflow_id,
            "name": self.name,
            "create_at": str(self.create_at), # "%Y-%m-%d %H:%M:%S.%f")
            "update_at": str(self.update_at),
            "configuration": json.loads(self.configuration),
            "description": self.description,
            "enable": self.enable,
        }

    def parse_dict(self, source):
        result = False

        attrs = [
            "workflow_id",
            "name",
            "create_at",
            "update_at",
            "configuration",
            "description",
            "enable",
        ]

        if hasattr(source, "__getitem__"):
            for attr in attrs:
                try:
                    setattr(self, attr, source[attr])
                except:
                    LOG.debug("some exception occured when extract %s attribute to object, i will discard it",
                        attr)
                    continue
            result = True
        else:
            LOG.debug("input param source does not have dict-like method, so i will do nothing at all!")
        return result

    def __repr__(self):
        return "workflow: %s" % self.to_dict()


BaseWorks = declarative_base()

class WorksTable(BaseWorks):
    __tablename__ = "works"

    id = Column(Integer, primary_key = True, autoincrement = True)
    work_id = Column(Text, unique = True, nullable = False, index = True)
    name = Column(Text, nullable = False, index = True)
    workflow_id = Column(Text, nullable = False, index = True)
    create_at = Column(DateTime, nullable = False, index = True)
    start_at = Column(DateTime, index = True)
    update_at = Column(DateTime, nullable = False, index = True)
    end_at = Column(DateTime, index = True)
    stage = Column(Text, nullable = False, index = True)
    status = Column(Text, index = True)
    input_data = Column(Text)
    configuration = Column(Text, nullable = False)
    result = Column(Text)

    @classmethod
    def init_engine_and_session(cls):
        cls.engine = create_engine('sqlite:///' + os.path.join(CONFIG["data_path"], "works.db"), echo = False)
        cls.session = sessionmaker(bind = cls.engine)
        return cls.engine, cls.session

    def to_dict(self):
        return {
            "id": self.id,
            "work_id": self.work_id,
            "name": self.name,
            "workflow_id": self.workflow_id,
            "create_at": str(self.create_at),
            "start_at": str(self.start_at) if self.start_at else self.start_at,
            "update_at": str(self.update_at),
            "end_at": str(self.end_at) if self.end_at else self.end_at,
            "stage": self.stage,
            "status": self.status,
            "input_data": json.loads(self.input_data),
            "configuration": json.loads(self.configuration),
            "result": json.loads(self.result),
        }

    def parse_dict(self, source):
        result = False

        attrs = [
            "work_id",
            "name",
            "workflow_id",
            "create_at",
            "start_at",
            "update_at",
            "end_at",
            "stage",
            "status",
            "input_data",
            "configuration",
            "result",
        ]

        if hasattr(source, "__getitem__"):
            for attr in attrs:
                try:
                    setattr(self, attr, source[attr])
                except:
                    LOG.debug("some exception occured when extract %s attribute to object, i will discard it",
                        attr)
                    continue
            result = True
        else:
            LOG.debug("input param source does not have dict-like method, so i will do nothing at all!")
        return result

    def __repr__(self):
        return "work: %s" % self.to_dict()


BaseSchedules = declarative_base()

class SchedulesTable(BaseSchedules):
    __tablename__ = "schedules"

    id = Column(Integer, primary_key = True, autoincrement = True)
    schedule_id = Column(Text, unique = True, nullable = False, index = True)
    schedule_name = Column(Text, nullable = False, index = True)
    source = Column(Text, nullable = False)
    source_id = Column(Text, nullable = False, index = True)
    create_at = Column(DateTime, nullable = False, index = True)
    update_at = Column(DateTime, nullable = False, index = True)
    input_data = Column(Text)
    minute = Column(Integer, nullable = True) # [0, 59]
    hour = Column(Integer, nullable = True) # [0, 23]
    day_of_month = Column(Integer, nullable = True) # [1, 31]
    day_of_week = Column(Integer, nullable = True) # [1, 7] (Sunday = 7)
    enable = Column(Boolean, nullable = False)

    @classmethod
    def init_engine_and_session(cls):
        cls.engine = create_engine('sqlite:///' + os.path.join(CONFIG["data_path"], "schedules.db"), echo = False)
        cls.session = sessionmaker(bind = cls.engine)
        return cls.engine, cls.session

    def to_dict(self):
        return {
            "id": self.id,
            "schedule_id": self.schedule_id,
            "schedule_name": self.schedule_name,
            "source": self.source,
            "source_id": self.source_id,
            "create_at": str(self.create_at),
            "update_at": str(self.update_at),
            "input_data": json.loads(self.input_data),
            "minute": self.minute,
            "hour": self.hour,
            "day_of_month": self.day_of_month,
            "day_of_week": self.day_of_week,
            "enable": self.enable,
        }

    def parse_dict(self, source):
        result = False

        attrs = [
            "schedule_id",
            "schedule_name",
            "source",
            "source_id",
            "create_at",
            "update_at",
            "input_data",
            "minute",
            "hour",
            "day_of_month",
            "day_of_week",
            "enable",
        ]

        if hasattr(source, "__getitem__"):
            for attr in attrs:
                try:
                    setattr(self, attr, source[attr])
                except:
                    LOG.debug("some exception occured when extract %s attribute to object, i will discard it",
                        attr)
                    continue
            result = True
        else:
            LOG.debug("input param source does not have dict-like method, so i will do nothing at all!")
        return result

    def __repr__(self):
        return "schedule: %s" % self.to_dict()


BaseServices = declarative_base()

class ServicesTable(BaseServices):
    __tablename__ = "services"

    id = Column(Integer, primary_key = True, autoincrement = True)
    service_id = Column(Text, unique = True, nullable = False, index = True)
    name = Column(Text, nullable = False, index = True)
    application_id = Column(Text, nullable = False, index = True)
    task_id = Column(Text, nullable = False, index = True)
    create_at = Column(DateTime, nullable = False, index = True)
    update_at = Column(DateTime, nullable = False, index = True)
    stage = Column(Text, nullable = False, index = True)
    status = Column(Text, index = True)
    input_data = Column(Text)
    description = Column(Text)
    signal = Column(Integer, nullable = True) # -9, -15
    enable = Column(Boolean, nullable = False)

    @classmethod
    def init_engine_and_session(cls):
        cls.engine = create_engine('sqlite:///' + os.path.join(CONFIG["data_path"], "services.db"), echo = False)
        cls.session = sessionmaker(bind = cls.engine)
        return cls.engine, cls.session

    def to_dict(self):
        return {
            "id": self.id,
            "service_id": self.service_id,
            "name": self.name,
            "application_id": self.application_id,
            "task_id": self.task_id,
            "create_at": str(self.create_at),
            "update_at": str(self.update_at),
            "stage": self.stage,
            "status": self.status,
            "input_data": json.loads(self.input_data),
            "description": self.description,
            "signal": self.signal,
            "enable": self.enable,
        }

    def parse_dict(self, source):
        result = False

        attrs = [
            "service_id",
            "name",
            "application_id",
            "task_id",
            "create_at",
            "update_at",
            "stage",
            "status",
            "input_data",
            "description",
            "signal",
            "enable",
        ]

        if hasattr(source, "__getitem__"):
            for attr in attrs:
                try:
                    setattr(self, attr, source[attr])
                except:
                    LOG.debug("some exception occured when extract %s attribute to object, i will discard it",
                        attr)
                    continue
            result = True
        else:
            LOG.debug("input param source does not have dict-like method, so i will do nothing at all!")
        return result

    def __repr__(self):
        return "service: %s" % self.to_dict()



# -*- coding: utf-8 -*-

import json
import time
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.manager.utils.scheduler import Scheduler
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.common import Errors
from litepipeline.version import __version__
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class AboutHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"message": "LitePipeline manager service"}
        self.write(result)
        self.finish()


class ClusterInfoHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK, "version": __version__}
        try:
            include = self.get_argument("include", "manager,nodes,actions")
            include = include.split(",")
            info = {}
            if "manager" in include:
                info["manager"] = CONFIG
                info["manager"]["version"] = __version__
                info["manager"]["total_action_slots"] = Connection.get_total_action_slots()
            if "nodes" in include:
                info["nodes"] = []
                info["number_of_nodes"] = 0
                for node in Connection.clients:
                    info["nodes"].append(node.info)
                    info["number_of_nodes"] += 1
            if "actions" in include:
                info["actions"] = Scheduler.instance().current_schedule_actions()
            result["info"] = info
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(json.dumps(result, sort_keys = True))
        self.finish()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import signal
import logging
import argparse
import platform

import tornado.ioloop
import tornado.httpserver
import tornado.web

from litepipeline.version import __version__
from litepipeline.node.handlers import info
from litepipeline.node.handlers import action
from litepipeline.node.handlers import task
from litepipeline.node.utils.registrant import Registrant
from litepipeline.node.utils.executor import Executor
from litepipeline.node.utils import common
from litepipeline.node.utils.apps_manager import ManagerClient as AppsManagerClient
from litepipeline.node.utils.workspace_manager import ManagerClient as WorkspaceManagerClient
from litepipeline.node.utils.persistent_config import PersistentConfig
from litepipeline.node.config import CONFIG, load_config
from litepipeline.node import logger

LOG = logging.getLogger(__name__)


class Application(tornado.web.Application):
    def __init__(self):
        handlers = [
            (r"/", info.AboutHandler),
            (r"/action/run", action.RunActionHandler),
            (r"/action/stop", action.StopActionHandler),
            (r"/action/cancel", action.CancelActionHandler),
            (r"/status/full", action.FullStatusHandler),
            (r"/workspace/pack", action.PackWorkspaceHandler),
            (r"/workspace/download", action.DownloadWorkspaceHandler),
            (r"/workspace/delete", task.DeleteTaskWorkspaceHandler),
        ]
        settings = dict(debug = False)
        tornado.web.Application.__init__(self, handlers, **settings)


def main():
    parser = argparse.ArgumentParser(prog = 'litenode')
    parser.add_argument("-c", "--config", required = True, help = "configuration file path")
    parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
    args = parser.parse_args()

    if args.config:
        success = load_config(args.config)
        if success:
            common.init_storage()
            logger.config_logging(file_name = "node.log",
                                  log_level = CONFIG["log_level"],
                                  dir_name = CONFIG["log_path"],
                                  day_rotate = False,
                                  when = "D",
                                  interval = 1,
                                  max_size = 20,
                                  backup_count = 5,
                                  console = True)
            LOG.info("service start")

            try:
                init_run = True
                config_file_path = os.path.join(CONFIG["data_path"], "configuration.json")
                if os.path.exists(config_file_path):
                    init_run = False
                C = PersistentConfig(config_file_path)
                if init_run:
                    C.from_dict(CONFIG)
                C.set("python3", sys.version)
                C.set("version", __version__)
                C.set("platform", platform.machine())

                node_registrant = Registrant(
                    C["manager_tcp_host"],
                    C["manager_tcp_port"],
                    C,
                    retry_interval = C["retry_interval"]
                )
                action_executor = Executor(C)

                http_server = tornado.httpserver.HTTPServer(Application())
                http_server.listen(C["http_port"], address = C["http_host"])
                # http_server.bind(C["http_port"], address = C["http_host"])
                common.Servers.HTTP_SERVER = http_server
                common.Servers.DB_SERVERS.append(action_executor)
                common.Servers.DB_SERVERS.append(AppsManagerClient())
                common.Servers.DB_SERVERS.append(WorkspaceManagerClient())
                tornado.ioloop.IOLoop.instance().add_callback(node_registrant.connect)
                signal.signal(signal.SIGTERM, common.sig_handler)
                signal.signal(signal.SIGINT, common.sig_handler)
                tornado.ioloop.IOLoop.instance().start()
            except Exception as e:
                LOG.exception(e)

            LOG.info("service end")
        else:
            print("failed to load configuration: %s" % args.config)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os
import time
import json
import logging
import shutil
import tarfile

import requests
from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, StreamBaseHandler
from litepipeline.manager.utils.app_manager import AppManager
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateApplicationHandler(StreamBaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            name = self.get_form_argument("name", "")
            description = self.get_form_argument("description", "")
            file_path = self.file_path.decode("utf-8")
            if name and os.path.exists(file_path) and os.path.isfile(file_path):
                file_name = os.path.split(file_path)[-1].lower()
                if ((CONFIG["app_store"].endswith("tar.gz") and file_name.endswith("tar.gz")) or
                    (CONFIG["app_store"].endswith("zip") and file_name.endswith("zip"))):
                    result["app_id"] = AppManager.instance().create(name, description, file_path)
                else:
                    LOG.warning("application wrong format")
                    Errors.set_result_error("AppWrongFormat", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        LOG.debug("deploy application package use: %ss", time.time() - self.start)
        self.write(result)
        self.finish()


class ListApplicationHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            app_id = self.get_argument("id", "")
            if app_id:
                filters["id"] = app_id
            LOG.debug("ListApplicationHandler offset: %s, limit: %s", offset, limit)
            r = AppManager.instance().list(offset, limit, filters = filters)
            result["apps"] = r["apps"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteApplicationHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_argument("app_id", "")
            if app_id:
                success = AppManager.instance().delete(app_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateApplicationHandler(StreamBaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_form_argument("app_id", "")
            name = self.get_form_argument("name", "")
            description = self.get_form_argument("description", "")
            LOG.debug("UpdateApplicationHandler app_id: %s, name: %s, description: %s", app_id, name, description)
            if app_id and AppManager.instance().info(app_id):
                success = AppManager.instance().update(app_id, name, description, self.file_path.decode("utf-8"))
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        LOG.debug("update application package use: %ss", time.time() - self.start)
        self.write(result)
        self.finish()


class InfoApplicationHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_argument("app_id", "")
            config = self.get_argument("config", "false")
            config = True if config.lower() == "true" else False
            if app_id:
                app_info = AppManager.instance().info(app_id)
                if app_info:
                    if config:
                        app_config = AppManager.instance().get_app_config(app_id, app_info["sha1"])
                        if app_config:
                            result["app_info"] = app_info
                            result["app_config"] = app_config
                        else:
                            Errors.set_result_error("OperationFailed", result)
                    else:
                        result["app_info"] = app_info
                elif app_info is None:
                    Errors.set_result_error("AppNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DownloadApplicationHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_argument("app_id", "")
            sha1 = self.get_argument("sha1", "")
            if app_id:
                app_info = AppManager.instance().info(app_id)
                if app_info:
                    if sha1 == "":
                        sha1 = app_info["sha1"]
                    f = AppManager.instance().open(app_id, sha1)
                    if f:
                        self.set_header('Content-Type', 'application/octet-stream')
                        if "app_store" in CONFIG:
                            if "tar.gz" in CONFIG["app_store"]:
                                self.set_header('Content-Disposition', 'attachment; filename=%s.tar.gz' % app_id)
                            elif "zip" in CONFIG["app_store"]:
                                self.set_header('Content-Disposition', 'attachment; filename=%s.zip' % app_id)
                            else:
                                self.set_header('Content-Disposition', 'attachment; filename=%s.tar.gz' % app_id)
                        else:
                            self.set_header('Content-Disposition', 'attachment; filename=%s.tar.gz' % app_id)
                        buf_size = 1024 * 1024
                        while True:
                            data = f.read(buf_size)
                            if not data:
                                break
                            self.write(data)
                            self.flush()
                            yield gen.moment
                        f.close()
                        self.finish()
                        return
                    else:
                        Errors.set_result_error("OperationFailed", result)
                elif app_info is None:
                    Errors.set_result_error("AppNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.set_status(400)
        self.write(result)
        self.finish()


class ApplicationHistoryListHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK, "app_histories": [], "total": 0}
        try:
            app_id = self.get_argument("app_id", "")
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            if app_id:
                app_info = AppManager.instance().info(app_id)
                if app_info:
                    app_histories = AppManager.instance().list_history(offset, limit, {"app_id": app_id})
                    if app_histories:
                        result["app_histories"] = app_histories["histories"]
                        result["total"] = app_histories["total"]
                    result["offset"] = offset
                    result["limit"] = limit
                elif app_info is None:
                    Errors.set_result_error("AppNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ApplicationHistoryInfoHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_argument("app_id", "")
            history_id = int(self.get_argument("history_id", "-1"))
            config = self.get_argument("config", "false")
            config = True if config.lower() == "true" else False
            if history_id != -1:
                app_history = AppManager.instance().info_history(history_id, app_id)
                if app_history:
                    if config:
                        app_config = AppManager.instance().get_app_config(app_id, app_history["sha1"])
                        if app_config:
                            result["history_info"] = app_history
                            result["history_config"] = app_config
                        else:
                            Errors.set_result_error("OperationFailed", result)
                    else:
                        result["history_info"] = app_history
                elif app_history is None:
                    Errors.set_result_error("AppHistoryNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ApplicationHistoryActivateHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            history_id = self.get_json_argument("history_id", "")
            app_id = self.get_json_argument("app_id", "")
            if history_id and app_id:
                success = AppManager.instance().activate_history(history_id, app_id = app_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ApplicationHistoryDeleteHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            app_id = self.get_argument("app_id", "")
            history_id = int(self.get_argument("history_id", "-1"))
            if history_id != -1 and app_id:
                success = AppManager.instance().delete_history(history_id, app_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


# -*- coding: utf-8 -*-

import json
import time
import urllib
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.manager.models.workflows import Workflows
from litepipeline.manager.models.works import Works
from litepipeline.manager.utils.scheduler import Scheduler
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, Stage, Status, Signal, splitall, JSONLoadError
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateWorkHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            name = self.get_json_argument("name", "")
            workflow_id = self.get_json_argument("workflow_id", "")
            input_data = self.get_json_argument("input_data", {})
            workflow = Workflows.instance().get(workflow_id)
            if workflow and name:
                if not isinstance(input_data, dict):
                    raise JSONLoadError("input_data must be dict type")
                if workflow["enable"]:
                    work_id = Works.instance().add(name, workflow_id, input_data = input_data, configuration = workflow["configuration"])
                    if work_id is not False:
                        result["work_id"] = work_id
                    else:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("WorkflowDisabled", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("CreateWorkHandler, name: %s, workflow_id: %s", name, workflow_id)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RunWorkHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            work_id = self.get_json_argument("work_id", "")
            if work_id:
                work = Works.instance().get(work_id)
                if work:
                    workflow = Workflows.instance().get(work["workflow_id"])
                    if workflow:
                        if workflow["enable"]:
                            if work["stage"] == Stage.finished:
                                success = Works.instance().update(work_id, {"stage": Stage.pending, "status": None})
                                if not success:
                                    Errors.set_result_error("OperationFailed", result)
                            else:
                                Errors.set_result_error("WorkStillRunning", result)
                        else:
                            Errors.set_result_error("WorkflowDisabled", result)
                    elif workflow is None:
                        Errors.set_result_error("WorkflowNotExists", result)
                    else:
                        Errors.set_result_error("OperationFailed", result)
                elif work is None:
                    Errors.set_result_error("WorkNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("RunWorkHandler, work_id: %s", work_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RecoverWorkHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            work_id = self.get_json_argument("work_id", "")
            if work_id:
                work = Works.instance().get(work_id)
                if work:
                    workflow = Workflows.instance().get(work["workflow_id"])
                    if workflow:
                        if workflow["enable"]:
                            if work["stage"] == Stage.finished:
                                if work["status"] != Status.success:
                                    success = Works.instance().update(work_id, {"stage": Stage.recovering, "status": None})
                                    if not success:
                                        Errors.set_result_error("OperationFailed", result)
                                else:
                                    Errors.set_result_error("WorkAlreadySuccess", result)
                            else:
                                Errors.set_result_error("WorkStillRunning", result)
                        else:
                            Errors.set_result_error("WorkflowDisabled", result)
                    elif workflow is None:
                        Errors.set_result_error("WorkflowNotExists", result)
                    else:
                        Errors.set_result_error("OperationFailed", result)
                elif work is None:
                    Errors.set_result_error("WorkNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("RecoverWorkHandler, work_id: %s", work_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class StopWorkHandler(BaseHandler): # kill -9 or -15
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            work_id = self.get_json_argument("work_id", "")
            signal = int(self.get_json_argument("signal", Signal.kill))
            if work_id and signal in (Signal.kill, Signal.terminate):
                work = Works.instance().get(work_id)
                if work:
                    Works.instance().update(work_id, {"stage": Stage.stopping})
                    if work["stage"] in (Stage.pending, Stage.recovering):
                        Works.instance().update(work_id, {"stage": Stage.finished, "status": Status.kill})
                    elif work["stage"] == Stage.running:
                        for app_name in work["result"]:
                            app = work["result"][app_name]
                            if "stage" not in app or ("stage" in app and app["stage"] != Stage.finished):
                                success = yield Scheduler.instance().stop_task(app["task_id"], signal)
                                if not success:
                                    Errors.set_result_error("OperationFailed", result)
                                    break
                    elif work["stage"] == Stage.finished:
                        Errors.set_result_error("WorkAlreadyFinished", result)
                elif Work is None:
                    Errors.set_result_error("WorkNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("StopWorkHandler, work_id: %s, signal: %s", work_id, signal)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteWorkHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            work_id = self.get_argument("work_id", "")
            if work_id:
                success = Works.instance().delete(work_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class InfoWorkHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            work_id = self.get_argument("work_id", "")
            if work_id:
                work_info = Works.instance().get(work_id)
                if work_info:
                    result["info"] = work_info
                elif work_info is None:
                    Errors.set_result_error("WorkNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListWorkHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            work_id = self.get_argument("work_id", "")
            if work_id:
                filters["work_id"] = work_id
            workflow_id = self.get_argument("workflow_id", "")
            if workflow_id:
                filters["workflow_id"] = workflow_id
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            stage = self.get_argument("stage", "")
            if stage:
                filters["stage"] = stage
            status = self.get_argument("status", "")
            if status:
                filters["status"] = status
            LOG.debug("ListWorkHandler offset: %s, limit: %s, filters: %s", offset, limit, filters)
            r = Works.instance().list(offset = offset, limit = limit, filters = filters)
            result["works"] = r["works"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CleanWorkHistoryHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {}
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import json
import time
import logging

from tornado import web
from tornado import gen

from litepipeline.node.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.node.config import CONFIG

LOG = logging.getLogger("__name__")


class AboutHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"message": "LitePipeline node service"}
        self.write(result)
        self.finish()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import signal
import logging
import argparse

import tornado.ioloop
import tornado.httpserver
import tornado.web

from litepipeline.version import __version__
from litepipeline.viewer.handlers import info, cluster, application, task, schedule, workflow, work, service
from litepipeline.viewer.utils import common
from litepipeline.viewer.config import CONFIG, load_config
from litepipeline.viewer import logger

LOG = logging.getLogger(__name__)

cwd = os.path.split(os.path.realpath(__file__))[0]


class Application(tornado.web.Application):
    def __init__(self):
        handlers = [
            (r"/", info.RedirectHandler),
            (r"/cluster", cluster.ClusterHandler),
            (r"/application", application.ApplicationHandler),
            (r"/application/(?P<app_id>.*)", application.ApplicationInfoHandler),
            (r"/task", task.TaskHandler),
            (r"/task/(?P<task_id>.*)", task.TaskInfoHandler),
            (r"/workflow", workflow.WorkflowHandler),
            (r"/work", work.WorkHandler),
            (r"/schedule", schedule.ScheduleHandler),
            (r"/service", service.ServiceHandler),
        ]
        settings = dict(
            debug = False,
            template_path = os.path.join(cwd, "templates"),
            static_path = os.path.join(cwd, "static")
        )
        tornado.web.Application.__init__(self, handlers, **settings)


def main():
    parser = argparse.ArgumentParser(prog = 'liteviewer')
    parser.add_argument("-c", "--config", required = True, help = "configuration file path")
    parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
    args = parser.parse_args()

    if args.config:
        success = load_config(args.config)
        if success:
            common.init_storage()
            logger.config_logging(file_name = "viewer.log",
                                  log_level = CONFIG["log_level"],
                                  dir_name = CONFIG["log_path"],
                                  day_rotate = False,
                                  when = "D",
                                  interval = 1,
                                  max_size = 20,
                                  backup_count = 5,
                                  console = True)
            LOG.info("service start")

            try:
                http_server = tornado.httpserver.HTTPServer(Application())
                http_server.listen(CONFIG["http_port"], address = CONFIG["http_host"])
                # http_server.bind(CONFIG["http_port"], address = CONFIG["http_host"])
                common.Servers.HTTP_SERVER = http_server
                signal.signal(signal.SIGTERM, common.sig_handler)
                signal.signal(signal.SIGINT, common.sig_handler)
                tornado.ioloop.IOLoop.instance().start()
            except Exception as e:
                LOG.exception(e)

            LOG.info("service end")
        else:
            print("failed to load configuration: %s" % args.config)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os
import json
import logging

from tornado import web
from tornado import gen

from litepipeline.node.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.node.utils.executor import Executor
from litepipeline.node.utils.common import Errors
from litepipeline.node.config import CONFIG

LOG = logging.getLogger("__name__")


class DeleteTaskWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_infos = self.get_json_argument("task_infos", [])
            failed_task_ids = yield Executor.instance().delete_task_workspace(task_infos)
            result["failed_task_ids"] = failed_task_ids
            LOG.info("DeleteTaskWorkspaceHandler, failed_task_ids: %s, task_infos: %s, data: %s", failed_task_ids, task_infos, self.json_data)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import os
import json
import logging
import time
import datetime
from copy import deepcopy

from tornado import ioloop
from tornado import gen

from litepipeline.manager.utils.scheduler import Scheduler
from litepipeline.manager.config import CONFIG
from litepipeline.manager import logger

LOG = logging.getLogger(__name__)


class Servers(object):
    HTTP_SERVER = None
    SERVERS = []
    TORNADO_INSTANCE = None


class StopService(object):
    _instance = None
    name = "stop_service"

    def __new__(cls, interval = 1):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.interval = interval
            cls._instance.services = [
                Scheduler.instance()
            ]
            cls._instance.running = False
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def stop(self):
        if not self.running:
            self.running = True
            self.periodic_crontab = ioloop.PeriodicCallback(
                self.stop_service,
                5 * 1000 # 30 seconds
            )
            self.periodic_crontab.start()
        else:
            LOG.warning("stop service already running")

    @gen.coroutine
    def stop_service(self):
        LOG.debug("stop_service")
        try:
            ready = 0
            for service in self.services:
                if service.name == "scheduler":
                    if not service.stop:
                        service.stop = True
                    else:
                        if len(service.tasks) > 0:
                            LOG.warning("some task still running: %s", len(service.tasks))
                        else:
                            LOG.info("all task stopped")
                            service.close()
                            ready += 1
                            break
            if len(self.services) == ready:
                for s in Servers.SERVERS:
                    s.close()
                    if hasattr(s, "name"):
                        LOG.info("Stop %s server!", s.name)
                    else:
                        LOG.info("Stop nameless server!")
                ioloop.IOLoop.current().add_callback_from_signal(self.shutdown)
        except Exception as e:
            LOG.exception(e)

    async def shutdown(self):
        LOG.info("Stopping Service(%s:%s)", CONFIG["http_host"], CONFIG["http_port"])
        if Servers.HTTP_SERVER:
            Servers.HTTP_SERVER.stop()
            LOG.info("Stop http server!")
        await gen.sleep(1)
        LOG.info("Will shutdown ...")
        ioloop.IOLoop.current().stop()


async def shutdown():
    s = StopService()
    s.stop()


def sig_handler(sig, frame):
    LOG.warning("sig_handler Caught signal: %s", sig)
    ioloop.IOLoop.current().add_callback_from_signal(shutdown)



# -*- coding: utf-8 -*-

import os
import sys
import time
import json
import logging

cwd = os.path.split(os.path.realpath(__file__))[0]
sys.path.insert(0, os.path.split(cwd)[0])

from litepipeline.manager.models.application_history import ApplicationHistory 
from litepipeline.manager import logger
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger(__name__)



if __name__ == "__main__":
    logger.config_logging(file_name = "test_sqlite_application_history.log",
                          log_level = "DEBUG",
                          dir_name = os.path.join(cwd, "logs"),
                          day_rotate = False,
                          when = "D",
                          interval = 1,
                          max_size = 20,
                          backup_count = 5,
                          console = True)

    LOG.debug("test start")
    
    try:
        CONFIG["data_path"] = cwd

        db = ApplicationHistory()

        LOG.debug("add: %s", db.add("id_1", "sha1_1", "description_1"))
        LOG.debug("add: %s", db.add("id_1", "sha1_2", "description_2"))
        LOG.debug("add: %s", db.add("id_1", "sha1_3", "description_3"))
        LOG.debug("add: %s", db.add("id_2", "sha1_4", "description_4"))
        LOG.debug("add: %s", db.add("id_2", "sha1_5", "description_5"))
        LOG.debug("add: %s", db.add("id_3", "sha1_6", "description_6"))

        LOG.debug("list: %s", json.dumps(db.list(), indent = 4))
        LOG.debug("list id_1: %s", json.dumps(db.list(filters = {"id": "id_1"}), indent = 4))
        LOG.debug("latest id_1: %s", json.dumps(db.get_latest("id_1"), indent = 4))

        LOG.debug("delete id_1: %s", db.delete_by_app_id("id_1"))
        LOG.debug("list id_1: %s", json.dumps(db.list(filters = {"id": "id_1"}), indent = 4))

        LOG.debug("delete 6: %s", db.delete(6))
        LOG.debug("list: %s", json.dumps(db.list(), indent = 4))

        db.close()
    except Exception as e:
        LOG.exception(e)

    LOG.debug("test end")



# -*- coding: utf-8 -*-

import os
import sys
import time
import json
import logging
import shutil
import tarfile
import signal
from pathlib import Path
import threading
from threading import Thread, Lock
from multiprocessing import Process, Queue, Pipe

from tornado import gen, locks
import requests

from litepipeline.node.utils.common import file_sha1sum, splitall, get_workspace_path
from litepipeline.node.config import CONFIG
from litepipeline.node import logger

LOG = logging.getLogger(__name__)


class Command(object):
    pack_workspace = "pack_workspace"
    force_pack_workspace = "force_pack_workspace"
    exit = "exit"


class TasksCache(object):
    tasks = {}
    tasks_lock = Lock()

    @classmethod
    def set(cls, workspace):
        cls.tasks_lock.acquire()
        if workspace not in cls.tasks:
            cls.tasks[workspace] = False
        cls.tasks_lock.release()

    @classmethod
    def exists(cls, workspace):
        result = False
        cls.tasks_lock.acquire()
        if workspace in cls.tasks:
            result = True
        cls.tasks_lock.release()
        return result

    @classmethod
    def get(cls):
        result = False
        cls.tasks_lock.acquire()
        try:
            for workspace in cls.tasks:
                if not cls.tasks[workspace]:
                    cls.tasks[workspace] = True
                    result = workspace
                    break
        except Exception as e:
            LOG.exception(e)
        cls.tasks_lock.release()
        return result

    @classmethod
    def remove(cls, workspace):
        cls.tasks_lock.acquire()
        del cls.tasks[workspace]
        cls.tasks_lock.release()


class StoppableThread(Thread):
    """
    Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition.
    """

    def __init__(self):
        super(StoppableThread, self).__init__()
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()


class WorkerThread(StoppableThread):
    def __init__(self, pid):
        StoppableThread.__init__(self)
        Thread.__init__(self)
        self.pid = pid

    def run(self):
        LOG = logging.getLogger("worker")
        LOG.info("Worker(%03d) start", self.pid)
        try:
            while not self.stopped():
                try:
                    workspace = TasksCache.get()
                    if workspace:
                        path_parts = splitall(workspace)
                        task_id = path_parts[-2]
                        action_name = path_parts[-1]
                        tar_workspace = os.path.join(CONFIG["data_path"], "tmp", "download", "%s.%s.tar.gz" % (task_id, action_name))
                        tar_workspace_tmp = os.path.join(CONFIG["data_path"], "tmp", "download", "%s.%s.tmp.tar.gz" % (task_id, action_name))
                        if os.path.exists(tar_workspace_tmp) and os.path.isfile(tar_workspace_tmp):
                            os.remove(tar_workspace_tmp)
                        if os.path.exists(tar_workspace) and os.path.isfile(tar_workspace):
                            os.remove(tar_workspace)
                        t = tarfile.open(tar_workspace_tmp, mode = "x:gz")
                        t.add(workspace, arcname = "%s.%s" % (task_id, action_name))
                        t.close()
                        os.rename(tar_workspace_tmp, tar_workspace)
                        TasksCache.remove(workspace)
                    else:
                        time.sleep(0.5)
                except Exception as e:
                    LOG.exception(e)
        except Exception as e:
            LOG.exception(e)
        LOG.info("Worker(%03d) exit", self.pid)


class Manager(Process):
    def __init__(self, pipe_client, worker_num):
        Process.__init__(self)
        self.pipe_client = pipe_client
        self.worker_num = worker_num

    def run(self):
        logger.config_logging(file_name = "workspace_manager.log",
                              log_level = "NOSET",
                              dir_name = CONFIG["log_path"],
                              when = "D",
                              interval = 1,
                              max_size = 20,
                              backup_count = 5,
                              console = True)
        LOG = logging.getLogger("manager")

        def sig_handler(sig, frame):
            LOG.warning("sig_handler Caught signal: %s", sig)

        LOG.info("Manager start")
        try:
            signal.signal(signal.SIGTERM, sig_handler)
            signal.signal(signal.SIGINT, sig_handler)

            threads = []
            for i in range(self.worker_num):
                t = WorkerThread(i)
                t.start()
                threads.append(t)

            while True:
                LOG.debug("Manager main loop")
                command, task_id, create_at, action_name = self.pipe_client.recv()
                if command == Command.pack_workspace:
                    ready = False
                    try:
                        LOG.debug("pack workspace, task_id: %s, action_name: %s", task_id, action_name)
                        workspace = get_workspace_path(create_at, task_id, action_name)
                        tar_workspace = os.path.join(CONFIG["data_path"], "tmp", "download", "%s.%s.tar.gz" % (task_id, action_name))
                        if os.path.exists(tar_workspace) and os.path.isfile(tar_workspace):
                            ready = True
                        # archive workspace
                        if not ready:
                            TasksCache.set(workspace)
                    except Exception as e:
                        LOG.exception(e)
                    self.pipe_client.send((command, ready))
                if command == Command.force_pack_workspace:
                    ready = False
                    try:
                        LOG.debug("force pack workspace, task_id: %s, action_name: %s", task_id, action_name)
                        workspace = get_workspace_path(create_at, task_id, action_name)
                        tar_workspace = os.path.join(CONFIG["data_path"], "tmp", "download", "%s.%s.tar.gz" % (task_id, action_name))
                        if os.path.exists(tar_workspace) and os.path.isdir(tar_workspace):
                            os.remove(tar_workspace)
                        # archive workspace
                        if not ready:
                            TasksCache.set(workspace)
                    except Exception as e:
                        LOG.exception(e)
                    self.pipe_client.send((command, ready))
                elif command == Command.exit:
                    for t in threads:
                        t.stop()
                    break
            for t in threads:
                t.join()
        except Exception as e:
            LOG.exception(e)
        LOG.info("Manager exit")


class ManagerClient(object):
    process_list = []
    process_dict = {}
    write_lock = locks.Lock()
    _instance = None

    def __init__(self, worker_num = 1):
        if ManagerClient._instance is None:
            self.worker_num = worker_num if worker_num > 0 else 1
            LOG.debug("ManagerClient, worker_num: %s", self.worker_num)
            pipe_master, pipe_client = Pipe()
            p = Manager(pipe_client, self.worker_num)
            p.daemon = True
            ManagerClient.process_list.append(p)
            ManagerClient.process_dict["manager"] = [p, pipe_master]
            p.start()
            ManagerClient._instance = self
        else:
            self.worker_num = ManagerClient._instance.worker_num

    @gen.coroutine
    def pack_workspace(self, task_id, create_at, action_name, force = False):
        result = False
        LOG.debug("start pack workspace, task_id: %s, create_at: %s, action_name: %s", task_id, create_at, action_name)
        with (yield ManagerClient.write_lock.acquire()):
            LOG.debug("get pack workspace lock, task_id: %s, action_name: %s", task_id, action_name)
            if force:
                ManagerClient.process_dict["manager"][1].send((Command.force_pack_workspace, task_id, create_at, action_name))
            else:
                ManagerClient.process_dict["manager"][1].send((Command.pack_workspace, task_id, create_at, action_name))
            LOG.debug("send pack workspace, task_id: %s, create_at: %s, action_name: %s, force: %s", task_id, create_at, action_name, force)
            while not ManagerClient.process_dict["manager"][1].poll():
                yield gen.moment
            LOG.debug("recv pack workspace, task_id: %s, action_name: %s", task_id, action_name)
            r = ManagerClient.process_dict["manager"][1].recv()
            LOG.debug("end pack workspace, task_id: %s, action_name: %s, r: %s", task_id, action_name, r)
        if r[1]:
            result = r[1]
        raise gen.Return(result)

    def close(self):
        try:
            LOG.info("close ManagerClient")
            ManagerClient.process_dict["manager"][1].send((Command.exit, None, None, None))
            for p in ManagerClient.process_list[1:]:
                p.terminate()
            for p in ManagerClient.process_list:
                while p.is_alive():
                    time.sleep(0.5)
                    LOG.debug("sleep 0.5 second")
            LOG.info("All Process Exit!")
        except Exception as e:
            LOG.exception(e)



# -*- coding: utf-8 -*-

import logging
from uuid import uuid4

import tornado.tcpserver
from tornado import gen
from tornado.ioloop import IOLoop
from tornado_discovery.connection import BaseConnection
from tornado_discovery.listener import BaseListener
from tornado_discovery.common import crc32sum, Command, Status, Message

LOG = logging.getLogger(__name__)


class Connection(BaseConnection):
    clients_dict = {}
    total_action_slots = 0

    def __init__(self, stream, address, scheduler):
        super(Connection, self).__init__(stream, address)
        self.scheduler = scheduler

    @classmethod
    def get_total_action_slots(cls):
        cls.total_action_slots = 0
        try:
            for node_id in cls.clients_dict:
                node = cls.clients_dict[node_id]
                if "action_slots" in node.info:
                    cls.total_action_slots += node.info["action_slots"]
        except Exception as e:
            LOG.exception(e)
        return cls.total_action_slots

    @gen.coroutine
    def _on_connect(self):
        try:
            while True:
                refuse_connect_flag = False
                data = yield self.read_message()
                send_data = {}
                # register
                if "command" in data and data["command"] == Command.register:
                    self.info = data["data"]
                    if self.info["http_host"] == "0.0.0.0":
                        self.info["http_host"] = self._address[0]
                    # no node_id
                    if "node_id" not in self.info or self.info["node_id"] == None:
                        node_id = str(uuid4())
                        send_data = {
                            "command": Command.register,
                            "data": {
                                "status": Status.success,
                                "message": Status.success,
                                "node_id": node_id,
                            }
                        }
                        self.info["node_id"] = node_id
                        self._status = Status.registered
                        if self.info["node_id"] not in Connection.clients_dict:
                            Connection.clients_dict[self.info["node_id"]] = self
                    # register with node_id
                    else:
                        if self.info["node_id"] not in Connection.clients_dict:
                            Connection.clients_dict[self.info["node_id"]] = self
                        send_data = {
                            "command": Command.register,
                            "data": {
                                "status": Status.success,
                                "message": Status.success,
                                "node_id": self.info["node_id"],
                            }
                        }
                        self._status = Status.registered
                    if "action_slots" in self.info:
                        Connection.total_action_slots += self.info["action_slots"]
                elif "command" in data and data["command"] == Command.heartbeat:
                    self.info = data["data"]
                    if self.info["http_host"] == "0.0.0.0":
                        self.info["http_host"] = self._address[0]
                    if self._status == Status.registered:
                        send_data = {
                            "command": Command.heartbeat,
                            "data": {
                                "status": Status.success,
                                "message": Status.success,
                            }
                        }
                        if self._heartbeat_timeout:
                            IOLoop.instance().remove_timeout(self._heartbeat_timeout)
                        self._heartbeat_timeout = IOLoop.instance().add_timeout(
                            IOLoop.time(IOLoop.instance()) + self.info["heartbeat_timeout"],
                            self._remove_connection
                        )
                    else:
                        send_data = {
                            "command": Command.heartbeat,
                            "data": {
                                "status": Status.failure,
                                "message": "invalid node_id: %s" % self.info["node_id"],
                            }
                        }
                        refuse_connect_flag = True
                else:
                    LOG.error("Client(%s) invaild message error: %s", self._address, data)
                    send_data = {
                        "command": Command.error,
                        "data": {
                            "status": Status.failure,
                            "message": "Unknown Command!",
                        }
                    }
                self.send_message(send_data, refuse_connect_flag = refuse_connect_flag)
        except tornado.iostream.StreamClosedError:
            LOG.info("Closed: %s", self._address)
        except Exception as e:
            LOG.exception(e)

    def _remove_connection(self):
        if self in BaseConnection.clients:
            if "node_id" in self.info and self.info["node_id"] in Connection.clients_dict:
                del Connection.clients_dict[self.info["node_id"]]
            BaseConnection.clients.remove(self)
        if "action_slots" in self.info:
            Connection.total_action_slots -= self.info["action_slots"]
        if "node_id" in self.info:
            self.scheduler.repending_running_actions(self.info["node_id"])
        self._stream.close()
        LOG.warning("Client(%s) node_id: %s heartbeat_timeout", self._address, self.info["node_id"])

    def _refuse_connect(self):
        if self._heartbeat_timeout:
            IOLoop.instance().remove_timeout(self._heartbeat_timeout)
        if self in BaseConnection.clients:
            if "node_id" in self.info and self.info["node_id"] in Connection.clients_dict:
                del Connection.clients_dict[self.info["node_id"]]
            BaseConnection.clients.remove(self)
        if "action_slots" in self.info:
            Connection.total_action_slots -= self.info["action_slots"]
        if "node_id" in self.info:
            self.scheduler.repending_running_actions(self.info["node_id"])
        self._stream.close()
        LOG.warning("Refuse(%s) node_id: %s connect", self._address, self.info["node_id"])

    def _on_close(self):
        if self._heartbeat_timeout:
            IOLoop.instance().remove_timeout(self._heartbeat_timeout)
        if self in BaseConnection.clients:
            if "node_id" in self.info and self.info["node_id"] in Connection.clients_dict:
                del Connection.clients_dict[self.info["node_id"]]
            BaseConnection.clients.remove(self)
        if "action_slots" in self.info:
            Connection.total_action_slots -= self.info["action_slots"]
        if "node_id" in self.info:
            self.scheduler.repending_running_actions(self.info["node_id"])
        self._stream.close()
        LOG.info("Client(%s) closed", self._address)


class DiscoveryListener(BaseListener):
    def __init__(self, connection_cls, scheduler, ssl_options = None, **kwargs):
        LOG.info("DiscoveryListener start")
        self.connection_cls = connection_cls
        self.scheduler = scheduler
        tornado.tcpserver.TCPServer.__init__(self, ssl_options = ssl_options, **kwargs)

    def handle_stream(self, stream, address):
        LOG.debug("Incoming connection from %r", address)
        self.connection_cls(stream, address, self.scheduler)



# -*- coding: utf-8 -*-

import json
import datetime
import logging
from uuid import uuid4

from litepipeline.manager.db.sqlite_interface import WorkflowsTable, NoResultFound
from litepipeline.manager.config import CONFIG


LOG = logging.getLogger(__name__)


class Workflows(object):
    _instance = None
    name = "workflows"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.table = WorkflowsTable
            engine, session = WorkflowsTable.init_engine_and_session()
            cls._instance.table.metadata.create_all(engine)
            cls._instance.session = session(autoflush = False, autocommit = False)
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def _new_id(self):
        return str(uuid4())

    def add(self, name, configuration, description = "", enable = False):
        result = False
        workflow_id = self._new_id()
        now = datetime.datetime.now()
        item = {
            "workflow_id": workflow_id,
            "name": name,
            "create_at": now,
            "update_at": now,
            "configuration": json.dumps(configuration),
            "description": description,
            "enable": enable,
        }

        row = self.table()
        row.parse_dict(item)
        try:
            self.session.add(row)
            self.session.commit()
            result = workflow_id
            LOG.debug("add workflow: %s", row)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def update(self, workflow_id, data):
        result = False
        try:
            now = datetime.datetime.now()
            data["update_at"] = now
            if "configuration" in data:
                data["configuration"] = json.dumps(data["configuration"])
            self.session.query(self.table).filter_by(workflow_id = workflow_id).update(data)
            self.session.commit()
            result = True
            LOG.debug("update workflow: %s, %s", workflow_id, data)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def delete(self, workflow_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(workflow_id = workflow_id).one()
            self.session.delete(row)
            self.session.commit()
            result = True
            LOG.debug("delete workflow: %s", row)
        except Exception as e:
            LOG.exception(e)
        return result

    def get(self, workflow_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(workflow_id = workflow_id).one()
            result = row.to_dict()
        except NoResultFound:
            result = None
        except Exception as e:
            LOG.exception(e)
        return result

    def parse_filters(self, filters):
        result = []
        try:
            if "id" in filters:
                result.append(self.table.workflow_id == filters["id"])
            if "name" in filters:
                result.append(self.table.name.like("%s" % filters["name"].replace("*", "%%")))
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset = 0, limit = 0, filters = {}):
        result = {"workflows": [], "total": 0}
        try:
            offset = 0 if offset < 0 else offset
            limit = 0 if limit < 0 else limit
            filters = self.parse_filters(filters)
            result["total"] = self.count(filters)
            if filters:
                if limit:
                    rows = self.session.query(self.table).filter(*filters).order_by(self.table.create_at.desc()).offset(offset).limit(limit)
                elif offset:
                    rows = self.session.query(self.table).filter(*filters).order_by(self.table.create_at.desc()).offset(offset)
                else:
                    rows = self.session.query(self.table).filter(*filters).order_by(self.table.create_at.desc())
            else:
                if limit:
                    rows = self.session.query(self.table).order_by(self.table.create_at.desc()).offset(offset).limit(limit)
                elif offset:
                    rows = self.session.query(self.table).order_by(self.table.create_at.desc()).offset(offset)
                else:
                    rows = self.session.query(self.table).order_by(self.table.create_at.desc())
            for row in rows:
                result["workflows"].append(row.to_dict())
        except Exception as e:
            LOG.exception(e)
        return result

    def count(self, filters):
        result = 0
        try:
            if filters:
                result = self.session.query(self.table).filter(*filters).count()
            else:
                result = self.session.query(self.table).count()
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        self.session.close()



# -*- coding: utf-8 -*-

import datetime
import logging
from uuid import uuid4

from litepipeline.manager.db.sqlite_interface import ApplicationHistoryTable, NoResultFound, and_, or_, not_
from litepipeline.manager.config import CONFIG


LOG = logging.getLogger(__name__)


class ApplicationHistory(object):
    _instance = None
    name = "application_history"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.table = ApplicationHistoryTable
            engine, session = ApplicationHistoryTable.init_engine_and_session()
            cls._instance.table.metadata.create_all(engine)
            cls._instance.session = session(autoflush = False, autocommit = False)
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def add(self, app_id, sha1, description = ""):
        result = False
        now = datetime.datetime.now()
        item = {
            "application_id": app_id,
            "create_at": now,
            "sha1": sha1,
            "description": description,
        }

        row = self.table()
        row.parse_dict(item)
        try:
            self.session.add(row)
            self.session.commit()
            result = row.id
            LOG.debug("add application history: %s", row)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def delete(self, history_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(id = history_id).one()
            self.session.delete(row)
            self.session.commit()
            result = row.to_dict()
            LOG.debug("delete application history: %s", row)
        except NoResultFound:
            result = None
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def delete_by_app_id(self, app_id):
        result = False
        try:
            self.session.query(self.table).filter_by(application_id = app_id).delete()
            self.session.commit()
            result = True
            LOG.debug("delete application[%s] all history", app_id)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def delete_by_history_id_app_id(self, history_id, app_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(id = history_id).filter_by(application_id = app_id).one()
            self.session.delete(row)
            self.session.commit()
            result = row.to_dict()
            LOG.debug("delete application[%s] history[%s]", app_id, history_id)
        except Exception as e:
            LOG.exception(e)
            self.session.rollback()
        return result

    def get(self, history_id, app_id = ""):
        result = False
        try:
            if app_id:
                row = self.session.query(self.table).filter_by(id = history_id).filter_by(application_id = app_id).one()
            else:
                row = self.session.query(self.table).filter_by(id = history_id).one()
            result = row.to_dict()
        except NoResultFound:
            result = None
        except Exception as e:
            LOG.exception(e)
        return result

    def get_latest(self, app_id):
        result = False
        try:
            row = self.session.query(self.table).filter_by(application_id = app_id).order_by(self.table.create_at.desc()).first()
            result = row.to_dict()
        except NoResultFound:
            result = None
        except Exception as e:
            LOG.exception(e)
        return result

    def parse_filters(self, filters):
        result = []
        try:
            if "app_id" in filters:
                result.append(self.table.application_id == filters["app_id"])
            if "sha1" in filters:
                result.append(self.table.sha1 == filters["sha1"])
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset = 0, limit = 0, filters = {}):
        result = {"histories": [], "total": 0}
        try:
            offset = 0 if offset < 0 else offset
            limit = 0 if limit < 0 else limit
            filters = self.parse_filters(filters)
            result["total"] = self.count(filters)
            if filters:
                if limit:
                    rows = self.session.query(self.table).filter(*filters).order_by(self.table.create_at.desc()).offset(offset).limit(limit)
                elif offset:
                    rows = self.session.query(self.table).filter(*filters).order_by(self.table.create_at.desc()).offset(offset)
                else:
                    rows = self.session.query(self.table).filter(*filters).order_by(self.table.create_at.desc())
            else:
                if limit:
                    rows = self.session.query(self.table).order_by(self.table.create_at.desc()).offset(offset).limit(limit)
                elif offset:
                    rows = self.session.query(self.table).order_by(self.table.create_at.desc()).offset(offset)
                else:
                    rows = self.session.query(self.table).order_by(self.table.create_at.desc())
            for row in rows:
                result["histories"].append(row.to_dict())
        except Exception as e:
            LOG.exception(e)
        return result

    def count(self, filters):
        result = 0
        try:
            if filters:
                result = self.session.query(self.table).filter(*filters).count()
            else:
                result = self.session.query(self.table).count()
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        self.session.close()



# -*- coding: utf-8 -*-

import os
import logging
import argparse

from sqlalchemy import Index

from litepipeline.version import __version__
from litepipeline.manager.db.sqlite_interface import ApplicationsTable
from litepipeline.manager.db.sqlite_interface import TasksTable
from litepipeline.manager.db.sqlite_interface import WorkflowsTable
from litepipeline.manager.db.sqlite_interface import WorksTable
from litepipeline.manager.db.sqlite_interface import SchedulesTable
from litepipeline.manager.config import CONFIG, load_config
from litepipeline.manager import logger

LOG = logging.getLogger(__name__)


application_name_index = Index('ix_applications_name', ApplicationsTable.name)

task_name_index = Index('ix_tasks_task_name', TasksTable.task_name)
task_start_at_index = Index('ix_tasks_start_at', TasksTable.start_at)
task_end_at_index = Index('ix_tasks_end_at', TasksTable.end_at)
task_status_index = Index('ix_tasks_status', TasksTable.status)

workflow_name_index = Index('ix_workflows_name', WorkflowsTable.name)

work_name_index = Index('ix_works_name', WorksTable.name)
work_start_at_index = Index('ix_works_start_at', WorksTable.start_at)
work_end_at_index = Index('ix_works_end_at', WorksTable.end_at)
work_status_index = Index('ix_works_status', WorksTable.status)

schedule_name_index = Index('ix_schedules_name', SchedulesTable.schedule_name)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog = 'migrate_database.py')
    parser.add_argument("-c", "--config", required = True, help = "configuration file path")
    parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
    args = parser.parse_args()
    if args.config:
        success = load_config(args.config)
        if success:
            logger.config_logging(file_name = "migrate_database.log",
                                  log_level = CONFIG["log_level"],
                                  dir_name = CONFIG["log_path"],
                                  day_rotate = False,
                                  when = "D",
                                  interval = 1,
                                  max_size = 20,
                                  backup_count = 5,
                                  console = True)

            LOG.info("migrate start")
            engine, _ = ApplicationsTable.init_engine_and_session()
            application_name_index.create(bind = engine)

            engine, _ = TasksTable.init_engine_and_session()
            task_name_index.create(bind = engine)
            task_start_at_index.create(bind = engine)
            task_end_at_index.create(bind = engine)
            task_status_index.create(bind = engine)

            engine, _ = WorkflowsTable.init_engine_and_session()
            workflow_name_index.create(bind = engine)

            engine, _ = WorksTable.init_engine_and_session()
            work_name_index.create(bind = engine)
            work_start_at_index.create(bind = engine)
            work_end_at_index.create(bind = engine)
            work_status_index.create(bind = engine)

            engine, _ = SchedulesTable.init_engine_and_session()
            schedule_name_index.create(bind = engine)

            LOG.info("migrate end")
        else:
            print("failed to load configuration: %s" % args.config)
    else:
        parser.print_help()


# -*- coding: utf-8 -*-

import os
import time
import json
import hashlib
import logging

from tornado import ioloop
from tornado import gen

from litedfs_client.client import LiteDFSClient

from litepipeline.manager.config import CONFIG

LOG = logging.getLogger(__name__)
LDFS = None


class LiteDFS(object):
    def __init__(self, host, port):
        self.client = LiteDFSClient(host, port)



# -*- coding: utf-8 -*-

import os
import io
import re
import json
import time
import logging
import tarfile

import requests
from tornado import web
from tornado import gen
from tornado import httpclient

from litepipeline.node.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.node.utils.executor import Executor
from litepipeline.node.utils.common import Errors, Signal, file_sha1sum, get_workspace_path
from litepipeline.node.config import CONFIG

LOG = logging.getLogger("__name__")


class RunActionHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            app_id = self.get_json_argument("app_id", "")
            if task_id and app_id:
                Executor.instance().push_action_with_counter(self.json_data)
            else:
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("RunActionHandler, task_id: %s, app_id: %s, data: %s", task_id, app_id, self.json_data)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class StopActionHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            name = self.get_json_argument("name", "")
            signal = int(self.get_json_argument("signal", Signal.kill))
            if task_id and name and signal in (Signal.kill, Signal.terminate):
                success = Executor.instance().stop_action(task_id, name, signal)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("StopActionHandler, task_id: %s, data: %s", task_id, self.json_data)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CancelActionHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            name = self.get_json_argument("name", "")
            if task_id and name:
                success = Executor.instance().stop_action(task_id, name, Signal.cancel)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("CancelActionHandler, task_id: %s, data: %s", task_id, self.json_data)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class FullStatusHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK, "full": True}
        try:
            full = Executor.instance().is_full()
            result["full"] = full
            LOG.debug("FullStatusHandler, full: %s", full)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class PackWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            create_at = self.get_json_argument("create_at", "")
            name = self.get_json_argument("name", "")
            force = self.get_json_argument("force", False)
            if task_id and create_at and name:
                ready = yield Executor.instance().pack_action_workspace(task_id, create_at, name, force)
                if not ready:
                    Errors.set_result_error("OperationRunning", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("PackWorkspaceHandler, task_id: %s, create_at: %s, name: %s, data: %s", task_id, create_at, name, self.json_data)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DownloadWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            task_id = self.get_argument("task_id", "")
            create_at = self.get_argument("create_at", "")
            name = self.get_argument("name", "")
            if task_id and create_at and name:
                workspace = get_workspace_path(create_at, task_id, name)
                tar_workspace = os.path.join(CONFIG["data_path"], "tmp", "download", "%s.%s.tar.gz" % (task_id, name))
                if os.path.exists(tar_workspace) and os.path.isfile(tar_workspace):
                    buf_size = 64 * 1024
                    self.set_header('Content-Type', 'application/octet-stream')
                    self.set_header('Content-Disposition', 'attachment; filename=%s.%s.tar.gz' % (task_id, name))
                    with open(tar_workspace, 'rb') as f:
                        while True:
                            data = f.read(buf_size)
                            if not data:
                                break
                            self.write(data)
                            self.flush()
                            yield gen.moment
                    self.finish()
                    return
                elif os.path.exists(workspace) and os.path.isdir(workspace):
                    Errors.set_result_error("WorkspaceNotPacked", result)
                else:
                    Errors.set_result_error("WorkspaceNotExists", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("DownloadWorkspaceHandler")
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import signal
import logging
import argparse

import tornado.ioloop
import tornado.httpserver
import tornado.web

from litepipeline.version import __version__
from litepipeline.manager.handlers import info
from litepipeline.manager.handlers import application
from litepipeline.manager.handlers import task
from litepipeline.manager.handlers import workflow
from litepipeline.manager.handlers import work
from litepipeline.manager.handlers import schedule
from litepipeline.manager.handlers import service
from litepipeline.manager.utils.listener import Connection
from litepipeline.manager.utils.listener import DiscoveryListener
from litepipeline.manager.models.applications import Applications
from litepipeline.manager.models.application_history import ApplicationHistory
from litepipeline.manager.models.tasks import Tasks
from litepipeline.manager.models.workflows import Workflows
from litepipeline.manager.models.works import Works
from litepipeline.manager.models.schedules import Schedules
from litepipeline.manager.models.services import Services
from litepipeline.manager.utils.app_manager import AppManager
from litepipeline.manager.utils.scheduler import Scheduler
from litepipeline.manager.utils import common
from litepipeline.manager.utils import stop_service
from litepipeline.manager.utils.litedfs import LDFS, LiteDFS
from litepipeline.manager.config import CONFIG, load_config
from litepipeline.manager import logger

LOG = logging.getLogger(__name__)


class Application(tornado.web.Application):
    def __init__(self):
        handlers = [
            (r"/", info.AboutHandler),
            (r"/cluster/info", info.ClusterInfoHandler),
            (r"/app/create", application.CreateApplicationHandler),
            (r"/app/list", application.ListApplicationHandler),
            (r"/app/delete", application.DeleteApplicationHandler),
            (r"/app/update", application.UpdateApplicationHandler),
            (r"/app/info", application.InfoApplicationHandler),
            (r"/app/download", application.DownloadApplicationHandler),
            (r"/app/history/list", application.ApplicationHistoryListHandler),
            (r"/app/history/info", application.ApplicationHistoryInfoHandler),
            (r"/app/history/activate", application.ApplicationHistoryActivateHandler),
            (r"/app/history/delete", application.ApplicationHistoryDeleteHandler),
            (r"/task/create", task.CreateTaskHandler),
            (r"/task/list", task.ListTaskHandler),
            (r"/task/info", task.InfoTaskHandler),
            (r"/task/delete", task.DeleteTaskHandler),
            (r"/task/rerun", task.RunTaskHandler),
            (r"/task/recover", task.RecoverTaskHandler),
            (r"/task/stop", task.StopTaskHandler),
            (r"/action/update", task.UpdateActionHandler),
            (r"/workspace/delete", task.DeleteTaskWorkspaceHandler),
            (r"/workspace/pack", task.PackTaskWorkspaceHandler),
            (r"/workspace/download", task.DownloadTaskWorkspaceHandler),
            (r"/workflow/create", workflow.CreateWorkflowHandler),
            (r"/workflow/list", workflow.ListWorkflowHandler),
            (r"/workflow/delete", workflow.DeleteWorkflowHandler),
            (r"/workflow/update", workflow.UpdateWorkflowHandler),
            (r"/workflow/info", workflow.InfoWorkflowHandler),
            (r"/work/create", work.CreateWorkHandler),
            (r"/work/list", work.ListWorkHandler),
            (r"/work/info", work.InfoWorkHandler),
            (r"/work/delete", work.DeleteWorkHandler),
            (r"/work/rerun", work.RunWorkHandler),
            (r"/work/recover", work.RecoverWorkHandler),
            (r"/work/stop", work.StopWorkHandler),
            (r"/schedule/create", schedule.CreateScheduleHandler),
            (r"/schedule/update", schedule.UpdateScheduleHandler),
            (r"/schedule/list", schedule.ListScheduleHandler),
            (r"/schedule/delete", schedule.DeleteScheduleHandler),
            (r"/schedule/info", schedule.InfoScheduleHandler),
            (r"/service/create", service.CreateServiceHandler),
            (r"/service/update", service.UpdateServiceHandler),
            (r"/service/list", service.ListServiceHandler),
            (r"/service/delete", service.DeleteServiceHandler),
            (r"/service/info", service.InfoServiceHandler),
        ]
        settings = dict(debug = False)
        tornado.web.Application.__init__(self, handlers, **settings)


def main():
    parser = argparse.ArgumentParser(prog = 'litemanager')
    parser.add_argument("-c", "--config", required = True, help = "configuration file path")
    parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
    args = parser.parse_args()

    if args.config:
        success = load_config(args.config)
        if success:
            common.init_storage()
            logger.config_logging(file_name = "manager.log",
                                  log_level = CONFIG["log_level"],
                                  dir_name = CONFIG["log_path"],
                                  day_rotate = False,
                                  when = "D",
                                  interval = 1,
                                  max_size = 20,
                                  backup_count = 5,
                                  console = True)

            LOG.info("service start")

            try:
                if CONFIG["ldfs_http_host"] and CONFIG["ldfs_http_port"]:
                    LDFS = LiteDFS(CONFIG["ldfs_http_host"], CONFIG["ldfs_http_port"])
                tasks_db = Tasks()
                applications_db = Applications()
                application_history_db = ApplicationHistory()
                workflows_db = Workflows()
                works_db = Works()
                schedules_db = Schedules()
                services_db = Services()
                app_manager = AppManager()
                task_scheduler = Scheduler(CONFIG["scheduler_interval"])
                http_server = tornado.httpserver.HTTPServer(
                    Application(),
                    max_buffer_size = CONFIG["max_buffer_size"],
                    chunk_size = 10 * 1024 * 1024
                )
                http_server.listen(CONFIG["http_port"], address = CONFIG["http_host"])
                # http_server.bind(CONFIG["http_port"], address = CONFIG["http_host"])
                listener = DiscoveryListener(Connection, task_scheduler)
                listener.listen(CONFIG["tcp_port"], CONFIG["tcp_host"])
                stop_service.Servers.HTTP_SERVER = http_server
                stop_service.Servers.SERVERS.append(task_scheduler)
                stop_service.Servers.SERVERS.append(applications_db)
                stop_service.Servers.SERVERS.append(application_history_db)
                stop_service.Servers.SERVERS.append(tasks_db)
                stop_service.Servers.SERVERS.append(workflows_db)
                stop_service.Servers.SERVERS.append(works_db)
                stop_service.Servers.SERVERS.append(schedules_db)
                stop_service.Servers.SERVERS.append(services_db)
                stop_service.Servers.SERVERS.append(app_manager)
                signal.signal(signal.SIGTERM, stop_service.sig_handler)
                signal.signal(signal.SIGINT, stop_service.sig_handler)
                tornado.ioloop.IOLoop.instance().start()
            except Exception as e:
                LOG.exception(e)

            LOG.info("service end")
        else:
            print("failed to load configuration: %s" % args.config)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import json
import time
import logging

from tornado import web
from tornado import gen

from litepipeline.viewer.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.viewer.config import CONFIG

LOG = logging.getLogger("__name__")


class ClusterHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        self.render(
            "cluster/cluster.html",
            current_nav = "cluster",
            manager_host = "%s:%s" % (
                CONFIG["manager_http_host"],
                CONFIG["manager_http_port"])
        )



# -*- coding: utf-8 -*-

import json
import time
import urllib
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.manager.models.tasks import Tasks
from litepipeline.manager.utils.app_manager import AppManager
from litepipeline.manager.utils.scheduler import Scheduler
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, Stage, Status, Signal, splitall, JSONLoadError
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateTaskHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_name = self.get_json_argument("task_name", "")
            app_id = self.get_json_argument("app_id", "")
            input_data = self.get_json_argument("input_data", {})
            if app_id and AppManager.instance().info(app_id) and task_name:
                if not isinstance(input_data, dict):
                    raise JSONLoadError("input_data must be dict type")
                task_id = Tasks.instance().add(task_name, app_id, input_data = input_data)
                if task_id is not False:
                    result["task_id"] = task_id
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("CreateTaskHandler, task_name: %s, app_id: %s", task_name, app_id)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RunTaskHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            if task_id:
                task = Tasks.instance().get(task_id)
                if task:
                    if task["stage"] == Stage.finished:
                        success = Tasks.instance().update(task_id, {"stage": Stage.pending, "status": None})
                        if not success:
                            Errors.set_result_error("OperationFailed", result)
                    else:
                        Errors.set_result_error("TaskStillRunning", result)
                elif task is None:
                    Errors.set_result_error("TaskNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("RunTaskHandler, task_id: %s", task_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RecoverTaskHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            if task_id:
                task = Tasks.instance().get(task_id)
                if task:
                    if task["stage"] == Stage.finished:
                        if task["status"] != Status.success:
                            success = Tasks.instance().update(task_id, {"stage": Stage.recovering, "status": None})
                            if not success:
                                Errors.set_result_error("OperationFailed", result)
                        else:
                            Errors.set_result_error("TaskAlreadySuccess", result)
                    else:
                        Errors.set_result_error("TaskStillRunning", result)
                elif task is None:
                    Errors.set_result_error("TaskNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("RecoverTaskHandler, task_id: %s", task_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class StopTaskHandler(BaseHandler): # kill -9 or -15
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            signal = int(self.get_json_argument("signal", Signal.kill))
            if task_id and signal in (Signal.kill, Signal.terminate):
                task = Tasks.instance().get(task_id)
                if task:
                    if task["stage"] != Stage.finished:
                        success = yield Scheduler.instance().stop_task(task_id, signal)
                        if not success:
                            Errors.set_result_error("OperationFailed", result)
                    else:
                        Errors.set_result_error("TaskAlreadyFinished", result)
                elif task is None:
                    Errors.set_result_error("TaskNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("StopTaskHandler, task_id: %s, signal: %s", task_id, signal)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteTaskHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            task_id = self.get_argument("task_id", "")
            if task_id:
                success = Tasks.instance().delete(task_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteTaskWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_ids = self.get_json_argument("task_ids", [])
            if isinstance(task_ids, list) and task_ids:
                result = yield Scheduler.instance().delete_task_workspace(task_ids)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("DeleteTaskWorkspaceHandler, task_ids: %s", task_ids)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class PackTaskWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            task_id = self.get_json_argument("task_id", "")
            name = self.get_json_argument("name", "")
            force = self.get_json_argument("force", False)
            if task_id and name:
                result = yield Scheduler.instance().pack_task_workspace(task_id, name, force)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("PackTaskWorkspaceHandler, task_id: %s, name: %s", task_id, name)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DownloadTaskWorkspaceHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        redirect_flag = False
        try:
            task_id = self.get_argument("task_id", "")
            name = self.get_argument("name", "")
            if task_id and name:
                r = yield Scheduler.instance().select_task_node_info(task_id, name)
                if "node_info" in r and r["node_info"] and r["task_info"]:
                    http_host = r["node_info"]["http_host"]
                    http_port = r["node_info"]["http_port"]
                    if http_host == "127.0.0.1":
                        host_parts = urllib.parse.urlsplit("//" + self.request.host)
                        http_host = host_parts.hostname
                    create_at = r["task_info"]["create_at"]
                    url = "http://%s:%s/workspace/download?task_id=%s&create_at=%s&name=%s" % (http_host, http_port, task_id, create_at, name)
                    self.redirect(url)
                    redirect_flag = True
                else:
                    result = r
            else:
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        if not redirect_flag:
            self.write(result)
            self.finish()


class InfoTaskHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            task_id = self.get_argument("task_id", "")
            if task_id:
                task_info = Tasks.instance().get(task_id)
                if task_info:
                    result["task_info"] = task_info
                    if task_info["stage"] == Stage.running:
                        result["task_running_info"] = Scheduler.instance().get_running_actions(task_id)
                elif task_info is None:
                    Errors.set_result_error("TaskNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListTaskHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            task_id = self.get_argument("task_id", "")
            if task_id:
                filters["task_id"] = task_id
            app_id = self.get_argument("app_id", "")
            if app_id:
                filters["app_id"] = app_id
            work_id = self.get_argument("work_id", "")
            if work_id:
                filters["work_id"] = work_id
            service_id = self.get_argument("service_id", "")
            if service_id:
                filters["service_id"] = service_id
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            stage = self.get_argument("stage", "")
            if stage:
                filters["stage"] = stage
            status = self.get_argument("status", "")
            if status:
                filters["status"] = status
            LOG.debug("ListTaskHandler offset: %s, limit: %s, filters: %s", offset, limit, filters)
            r = Tasks.instance().list(offset = offset, limit = limit, filters = filters)
            result["tasks"] = r["tasks"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CleanTaskHistoryHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {}
        self.write(result)
        self.finish()


class UpdateActionHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            name = self.get_json_argument("name", "")
            task_id = self.get_json_argument("task_id", "")
            stage = self.get_json_argument("stage", "")
            if task_id and name and stage:
                if stage == Stage.running:
                    Scheduler.instance().update_running_action(self.json_data)
                elif stage == Stage.finished:
                    Scheduler.instance().update_finish_action(self.json_data)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("UpdateActionHandler, action_name: %s, task_id: %s", name, task_id)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-
'''
Created on 2013-10-26 21:29
@summary:  import yaml configuration
@author: YangHaitao
''' 
try:
    import yaml
except ImportError:
    raise ImportError("Config module requires pyYAML package, please check if pyYAML is installed!")

from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

import os

cwd = os.path.split(os.path.realpath(__file__))[0]

CONFIG = {}

def load_config(config_file_path):
    result = False
    try:
        if config_file_path and os.path.exists(config_file_path) and os.path.isfile(config_file_path):
            s = open(config_file_path, "r")
            local_config = load(stream = s, Loader = Loader)
            CONFIG.update(local_config)
            s.close()
            if "app_path" not in CONFIG:
                CONFIG["app_path"] = cwd
            result = True
    except Exception as e:
        print(e)
    return result

if __name__ == "__main__":
    print ("CONFIG: %s" % CONFIG)



# -*- coding: utf-8 -*-

import os
import io
import time
import json
import logging
import shutil
import tarfile
import zipfile

from litedfs_client.client import LiteDFSClient

from litepipeline.manager.models.applications import Applications
from litepipeline.manager.models.application_history import ApplicationHistory
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, splitall
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class AppManagerBase(object):
    def __init__(self):
        pass

    def create(self, name, description, source_path):
        pass

    def update(self, app_id, name, description, source_path):
        pass

    def list(self, offset, limit, filters = {}):
        pass

    def list_history(self, offset, limit, filters = {}):
        pass

    def info(self, app_id):
        pass

    def info_history(self, history_id, app_id = ""):
        pass

    def activate_history(self, history_id, app_id = ""):
        pass

    def delete(self, app_id):
        pass

    def delete_history(self, history_id, app_id):
        pass

    def get_app_config(self, app_id, sha1):
        pass

    def open(self, app_id, sha1):
        pass

    def close(self):
        pass


class AppLocalTarGzManager(AppManagerBase):
    _instance = None
    name = "AppLocalTarGzManager"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.root_path = CONFIG["data_path"]
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def make_app_path(self, app_id):
        return os.path.join(self.root_path, "applications", app_id[:2], app_id[2:4], app_id)

    def make_app_version_path(self, app_id, sha1):
        return os.path.join(self.make_app_path(app_id), sha1)

    def create(self, name, description, source_path):
        sha1 = file_sha1sum(source_path)
        LOG.debug("sha1: %s, %s", sha1, type(sha1))
        app_id = Applications.instance().add(name, sha1, description = description)
        app_path = self.make_app_version_path(app_id, sha1)
        if os.path.exists(app_path):
            shutil.rmtree(app_path)
        os.makedirs(app_path)
        shutil.copy2(source_path, os.path.join(app_path, "app.tar.gz"))
        os.remove(source_path)
        if os.path.exists(os.path.join(app_path, "app")):
            shutil.rmtree(os.path.join(app_path, "app"))
        t = tarfile.open(os.path.join(app_path, "app.tar.gz"), "r")
        t.extractall(app_path)
        path_parts = splitall(t.getnames()[0])
        tar_root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
        t.close()
        os.rename(os.path.join(app_path, tar_root_name), os.path.join(app_path, "app"))
        ApplicationHistory.instance().add(app_id, sha1, description = description)
        return app_id

    def update(self, app_id, name, description, source_path):
        result = True
        try:
            data = {}
            need_update = False
            if name:
                data["name"] = name
            if description:
                data["description"] = description
            if os.path.exists(source_path) and os.path.isfile(source_path):
                sha1 = file_sha1sum(source_path)
                data["sha1"] = sha1
                LOG.debug("sha1: %s, %s", sha1, type(sha1))
                app_path = self.make_app_version_path(app_id, sha1)
                if os.path.exists(app_path):
                    shutil.rmtree(app_path)
                os.makedirs(app_path)
                shutil.copy2(source_path, os.path.join(app_path, "app.tar.gz"))
                os.remove(source_path)
                if os.path.exists(os.path.join(app_path, "app")):
                    shutil.rmtree(os.path.join(app_path, "app"))
                t = tarfile.open(os.path.join(app_path, "app.tar.gz"), "r")
                t.extractall(app_path)
                tar_root_name = splitall(t.getnames()[0])[0]
                os.rename(os.path.join(app_path, tar_root_name), os.path.join(app_path, "app"))
                need_update = True
            if data or need_update:
                success = Applications.instance().update(app_id, data)
                if success:
                    if "sha1" in data:
                        description = "" if "description" not in data else data["description"]
                        ApplicationHistory.instance().add(app_id, data["sha1"], description = description)
                else:
                    result = False
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset, limit, filters = {}):
        return Applications.instance().list(offset = offset, limit = limit, filters = filters)

    def list_history(self, offset, limit, filters = {}):
        return ApplicationHistory.instance().list(offset = offset, limit = limit, filters = filters)

    def info(self, app_id):
        return Applications.instance().get(app_id)

    def info_history(self, history_id, app_id = ""):
        return ApplicationHistory.instance().get(history_id, app_id = app_id)

    def activate_history(self, history_id, app_id = ""):
        result = False
        try:
            if app_id:
                history = self.info_history(history_id, app_id = app_id)
                if history:
                    data = {"sha1": history["sha1"]}
                    if history["description"]:
                        data["description"] = history["description"]
                    success = Applications.instance().update(app_id, data)
                    if success:
                        result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete(self, app_id):
        result = False
        try:
            success = Applications.instance().delete(app_id)
            if success:
                ApplicationHistory.instance().delete_by_app_id(app_id)
                app_path = self.make_app_path(app_id)
                if os.path.exists(app_path):
                    shutil.rmtree(app_path)
                    LOG.debug("remove directory: %s", app_path)
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_history(self, history_id, app_id):
        result = False
        try:
            history = ApplicationHistory.instance().delete_by_history_id_app_id(history_id, app_id)
            if history and history is not None:
                filters = ApplicationHistory.instance().parse_filters({"app_id": history["application_id"], "sha1": history["sha1"]})
                num = ApplicationHistory.instance().count(filters)
                if num == 0:
                    app_path = self.make_app_version_path(history["application_id"], history["sha1"])
                    if os.path.exists(app_path):
                        shutil.rmtree(app_path)
                        LOG.debug("remove directory: %s", app_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def get_app_config(self, app_id, sha1):
        result = False
        try:
            app_config_path = os.path.join(self.make_app_version_path(app_id, sha1), "app", "configuration.json")
            if os.path.exists(app_config_path):
                fp = open(app_config_path, "r")
                result = json.loads(fp.read())
                fp.close()
        except Exception as e:
            LOG.exception(e)
        return result

    def open(self, app_id, sha1):
        result = False
        try:
            app_path = os.path.join(self.make_app_version_path(app_id, sha1), "app.tar.gz")
            if os.path.exists(app_path) and os.path.isfile(app_path):
                result = open(app_path, "rb")
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        pass


class AppLocalZipManager(AppManagerBase):
    _instance = None
    name = "AppLocalZipManager"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.root_path = CONFIG["data_path"]
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def make_app_path(self, app_id):
        return os.path.join(self.root_path, "applications", app_id[:2], app_id[2:4], app_id)

    def make_app_version_path(self, app_id, sha1):
        return os.path.join(self.make_app_path(app_id), sha1)

    def create(self, name, description, source_path):
        sha1 = file_sha1sum(source_path)
        LOG.debug("sha1: %s, %s", sha1, type(sha1))
        app_id = Applications.instance().add(name, sha1, description = description)
        app_path = self.make_app_version_path(app_id, sha1)
        if os.path.exists(app_path):
            shutil.rmtree(app_path)
        os.makedirs(app_path)
        shutil.copy2(source_path, os.path.join(app_path, "app.zip"))
        os.remove(source_path)
        ApplicationHistory.instance().add(app_id, sha1, description = description)
        return app_id

    def update(self, app_id, name, description, source_path):
        result = True
        try:
            data = {}
            need_update = False
            if name:
                data["name"] = name
            if description:
                data["description"] = description
            if os.path.exists(source_path) and os.path.isfile(source_path):
                sha1 = file_sha1sum(source_path)
                data["sha1"] = sha1
                LOG.debug("sha1: %s, %s", sha1, type(sha1))
                app_path = self.make_app_version_path(app_id, sha1)
                if os.path.exists(app_path):
                    shutil.rmtree(app_path)
                os.makedirs(app_path)
                shutil.copy2(source_path, os.path.join(app_path, "app.zip"))
                os.remove(source_path)
                need_update = True
            if data or need_update:
                success = Applications.instance().update(app_id, data)
                if success:
                    if "sha1" in data:
                        description = "" if "description" not in data else data["description"]
                        ApplicationHistory.instance().add(app_id, data["sha1"], description = description)
                else:
                    result = False
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset, limit, filters = {}):
        return Applications.instance().list(offset = offset, limit = limit, filters = filters)

    def list_history(self, offset, limit, filters = {}):
        return ApplicationHistory.instance().list(offset = offset, limit = limit, filters = filters)

    def info(self, app_id):
        return Applications.instance().get(app_id)

    def info_history(self, history_id, app_id = ""):
        return ApplicationHistory.instance().get(history_id, app_id = app_id)

    def activate_history(self, history_id, app_id = ""):
        result = False
        try:
            if app_id:
                history = self.info_history(history_id, app_id = app_id)
                if history:
                    data = {"sha1": history["sha1"]}
                    if history["description"]:
                        data["description"] = history["description"]
                    success = Applications.instance().update(app_id, data)
                    if success:
                        result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete(self, app_id):
        result = False
        try:
            success = Applications.instance().delete(app_id)
            if success:
                ApplicationHistory.instance().delete_by_app_id(app_id)
                app_path = self.make_app_path(app_id)
                if os.path.exists(app_path):
                    shutil.rmtree(app_path)
                    LOG.debug("remove directory: %s", app_path)
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_history(self, history_id, app_id):
        result = False
        try:
            history = ApplicationHistory.instance().delete_by_history_id_app_id(history_id, app_id)
            if history and history is not None:
                filters = ApplicationHistory.instance().parse_filters({"app_id": history["application_id"], "sha1": history["sha1"]})
                num = ApplicationHistory.instance().count(filters)
                if num == 0:
                    app_path = self.make_app_version_path(history["application_id"], history["sha1"])
                    if os.path.exists(app_path):
                        shutil.rmtree(app_path)
                        LOG.debug("remove directory: %s", app_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def get_app_config(self, app_id, sha1):
        result = False
        try:
            app_path = os.path.join(self.make_app_version_path(app_id, sha1), "app.zip")
            if os.path.exists(app_path):
                z = zipfile.ZipFile(app_path)
                path_parts = splitall(z.namelist()[0])
                root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
                app_config_path = os.path.join(root_name, "configuration.json")
                result = json.loads(z.read(app_config_path).decode("utf-8"))
                z.close()
        except Exception as e:
            LOG.exception(e)
        return result

    def open(self, app_id, sha1):
        result = False
        try:
            app_path = os.path.join(self.make_app_version_path(app_id, sha1), "app.zip")
            if os.path.exists(app_path) and os.path.isfile(app_path):
                result = open(app_path, "rb")
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        pass


class AppLDFSZipManager(AppManagerBase):
    _instance = None
    name = "AppLocalZipManager"

    def __new__(cls):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.root_path = os.path.join("/litemanager_%s:%s" % (CONFIG["http_host"], CONFIG["http_port"]), "data")
            cls._instance.ldfs = LiteDFSClient(CONFIG["ldfs_http_host"], CONFIG["ldfs_http_port"])
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def make_app_path(self, app_id):
        return os.path.join(self.root_path, "applications", app_id[:2], app_id[2:4], app_id)

    def make_app_version_path(self, app_id, sha1):
        return os.path.join(self.make_app_path(app_id), sha1)

    def create(self, name, description, source_path):
        sha1 = file_sha1sum(source_path)
        LOG.debug("sha1: %s, %s", sha1, type(sha1))
        app_id = Applications.instance().add(name, sha1, description = description)
        app_path = self.make_app_version_path(app_id, sha1)
        self.ldfs.delete_directory(app_path)
        self.ldfs.create_file(source_path, os.path.join(app_path, "app.zip"), replica = 1)
        os.remove(source_path)
        ApplicationHistory.instance().add(app_id, sha1, description = description)
        return app_id

    def update(self, app_id, name, description, source_path):
        result = True
        try:
            data = {}
            need_update = False
            if name:
                data["name"] = name
            if description:
                data["description"] = description
            if os.path.exists(source_path) and os.path.isfile(source_path):
                sha1 = file_sha1sum(source_path)
                data["sha1"] = sha1
                LOG.debug("sha1: %s, %s", sha1, type(sha1))
                app_path = self.make_app_version_path(app_id, sha1)
                self.ldfs.delete_directory(app_path)
                self.ldfs.create_file(source_path, os.path.join(app_path, "app.zip"), replica = 1)
                os.remove(source_path)
                need_update = True
            if data or need_update:
                success = Applications.instance().update(app_id, data)
                if success:
                    if "sha1" in data:
                        description = "" if "description" not in data else data["description"]
                        ApplicationHistory.instance().add(app_id, data["sha1"], description = description)
                else:
                    result = False
        except Exception as e:
            LOG.exception(e)
        return result

    def list(self, offset, limit, filters = {}):
        return Applications.instance().list(offset = offset, limit = limit, filters = filters)

    def list_history(self, offset, limit, filters = {}):
        return ApplicationHistory.instance().list(offset = offset, limit = limit, filters = filters)

    def info(self, app_id):
        return Applications.instance().get(app_id)

    def info_history(self, history_id, app_id = ""):
        return ApplicationHistory.instance().get(history_id, app_id = app_id)

    def activate_history(self, history_id, app_id = ""):
        result = False
        try:
            if app_id:
                history = self.info_history(history_id, app_id = app_id)
                if history:
                    data = {"sha1": history["sha1"]}
                    if history["description"]:
                        data["description"] = history["description"]
                    success = Applications.instance().update(app_id, data)
                    if success:
                        result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete(self, app_id):
        result = False
        try:
            success = Applications.instance().delete(app_id)
            if success:
                ApplicationHistory.instance().delete_by_app_id(app_id)
                app_path = self.make_app_path(app_id)
                self.ldfs.delete_directory(app_path)
                LOG.debug("remove ldfs directory: %s", app_path)
                result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_history(self, history_id, app_id):
        result = False
        try:
            history = ApplicationHistory.instance().delete_by_history_id_app_id(history_id, app_id)
            if history and history is not None:
                filters = ApplicationHistory.instance().parse_filters({"app_id": history["application_id"], "sha1": history["sha1"]})
                num = ApplicationHistory.instance().count(filters)
                if num == 0:
                    app_path = self.make_app_version_path(history["application_id"], history["sha1"])
                    self.ldfs.delete_directory(app_path)
                    LOG.debug("remove ldfs directory: %s", app_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def get_app_config(self, app_id, sha1):
        result = False
        try:
            app_path = os.path.join(self.make_app_version_path(app_id, sha1), "app.zip")
            remote_file = self.ldfs.open_remote_file(app_path)
            if remote_file:
                z = zipfile.ZipFile(remote_file)
                path_parts = splitall(z.namelist()[0])
                root_name = path_parts[1] if path_parts[0] == "." else path_parts[0]
                app_config_path = os.path.join(root_name, "configuration.json")
                result = json.loads(z.read(app_config_path).decode("utf-8"))
                z.close()
        except Exception as e:
            LOG.exception(e)
        return result

    def open(self, app_id, sha1):
        result = False
        try:
            app_path = os.path.join(self.make_app_version_path(app_id, sha1), "app.zip")
            remote_file = self.ldfs.open_remote_file(app_path)
            if remote_file:
                result = io.BufferedReader(remote_file, buffer_size = 1024 * 1024)
        except Exception as e:
            LOG.exception(e)
        return result

    def close(self):
        pass


class AppManager(object):
    _instance = None
    name = "AppManager"

    def __new__(cls):
        if not cls._instance:
            if "app_store" in CONFIG:
                if CONFIG["app_store"] == "local.tar.gz":
                    cls._instance = AppLocalTarGzManager()
                elif CONFIG["app_store"] == "local.zip":
                    cls._instance = AppLocalZipManager()
                elif CONFIG["app_store"] == "ldfs.zip":
                    cls._instance = AppLDFSZipManager()
                else:
                    cls._instance = AppLocalZipManager()
            else:
                cls._instance = AppLocalZipManager()
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance



# -*- coding: utf-8 -*-

import os
import json
import logging
import tarfile
import shutil
import datetime
import subprocess
from pathlib import Path

from tornado.httpclient import AsyncHTTPClient, HTTPRequest
import tornado.ioloop
import tornado.web
from tornado import gen
import docker

from litepipeline.node.utils.common import Errors, Stage, Status, Signal, file_sha1sum, splitall, get_workspace_path
from litepipeline.node.utils.apps_manager import ManagerClient as AppsManagerClient
from litepipeline.node.utils.workspace_manager import ManagerClient as WorkspaceManagerClient
from litepipeline.node.utils.registrant import Registrant
from litepipeline.node.config import CONFIG
from litepipeline.node import logger

LOG = logging.getLogger(__name__)


class Executor(object):
    _instance = None

    def __new__(cls, config):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.config = config
            cls._instance.interval = config["executor_interval"]
            cls._instance.ioloop_service()
            cls._instance.running_actions = []
            cls._instance.actions_counter = 0
            cls._instance.async_client = AsyncHTTPClient()
            cls._instance.apps_manager = AppsManagerClient()
            cls._instance.workspace_manager = WorkspaceManagerClient()
            cls._instance.docker_client = None
            if CONFIG["docker_support"]:
                cls._instance.docker_client = docker.from_env()
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def ioloop_service(self):
        self.periodic_execute = tornado.ioloop.PeriodicCallback(
            self.execute_service, 
            self.interval * 1000
        )
        self.periodic_execute.start()

    def push_action_with_counter(self, action):
        self.running_actions.append(action)
        self.actions_counter += 1

    def push_action(self, action):
        self.running_actions.append(action)

    def pop_action(self):
        result = None
        try:
            result = self.running_actions.pop(0)
        except IndexError:
            LOG.debug("no more action to execute")
        return result

    @gen.coroutine
    def delete_task_workspace(self, task_infos):
        result = []
        running_task_ids = []
        try:
            for action in self.running_actions:
                running_task_ids.append(action["task_id"])
            for task_info in task_infos:
                task_id = task_info["task_id"]
                create_at = task_info["create_at"]
                if task_id in running_task_ids:
                    result.append(task_id)
                else:
                    workspace = get_workspace_path(create_at, task_id)
                    if os.path.exists(workspace):
                        shutil.rmtree(workspace)
                        LOG.info("delete workspace: %s", workspace)
                        yield gen.moment
        except Exception as e:
            LOG.exception(e)
        return result

    @gen.coroutine
    def pack_action_workspace(self, task_id, create_at, action_name, force = False):
        result = False
        try:
            result = yield self.workspace_manager.pack_workspace(task_id, create_at, action_name, force)
        except Exception as e:
            LOG.exception(e)
        return result

    def is_full(self):
        result = True
        try:
            LOG.debug("is_full, actions_counter: %s", self.actions_counter)
            result = self.actions_counter >= CONFIG["action_slots"]
        except Exception as e:
            LOG.exception(e)
        return result

    def current_running_actions(self):
        result = []
        try:
            result = self.running_actions
        except Exception as e:
            LOG.exception(e)
        return result

    def stop_action(self, task_id, action_name, signal):
        result = False
        try:
            for action in self.running_actions:
                if action["task_id"] == task_id and action["name"] == action_name:
                    action["signal"] = signal
                    break
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def update_heartbeat_data(self):
        try:
            data = {"actions_counter": self.actions_counter, "running_actions": []}
            for action in self.running_actions:
                data["running_actions"].append(
                    {
                        "task_id": action["task_id"],
                        "action_name": action["name"],
                        "update_at": str(action["update_at"]) if "update_at" in action else None,
                    }
                )
            Registrant.instance().update_heartbeat_data(data = data)
        except Exception as e:
            LOG.exception(e)

    @gen.coroutine
    def execute_service(self):
        LOG.debug("execute_service")
        try:
            action = self.pop_action()
            if action:
                name = action["name"]
                task_id = action["task_id"]
                app_id = action["app_id"]
                sha1 = action["app_sha1"]
                create_at = action["task_create_at"]
                workspace = get_workspace_path(create_at, task_id, name)
                if not os.path.exists(workspace):
                    os.makedirs(workspace)
                workspace = str(Path(workspace).resolve())
                # action not running yet
                if "process" not in action:
                    app_ready = yield self.apps_manager.check_app(app_id)
                    LOG.debug("execute action: %s, app_ready: %s", action, app_ready)
                    if app_ready is True: # start run app ready action
                        app_base_path = os.path.join(CONFIG["data_path"], "applications", app_id[:2], app_id[2:4], app_id)
                        app_path = os.path.join(app_base_path, "app")
                        app_path = str(Path(app_path).resolve())
                        LOG.debug("execute application[%s][%s]", app_path, name)
                        input_data = {"task_id": task_id, "workspace": workspace}
                        if "input_data" in action:
                            input_data.update(action["input_data"])
                        if not os.path.exists(workspace):
                            os.makedirs(workspace)
                        fp = open(os.path.join(workspace, "input.data"), "w")
                        fp.write(json.dumps(input_data))
                        fp.close()
                        main_path = action["main"]
                        if CONFIG["docker_support"] and "docker" in action and action["docker"] and "docker_registry" in action and action["docker_registry"]:
                            if "env" in action:
                                venv_path = os.path.join(action["env"], "bin", "activate")
                                cmd = "bash -c \"cd /opt/app && source '%s' && exec %s /opt/workspace --nodaemon\"" % (venv_path, main_path)
                            else:
                                cmd = "bash -c \"cd /opt/app && exec %s /opt/workspace --nodaemon\"" % (main_path, )
                            docker_args = {}
                            if "args" in action["docker"] and action["docker"]["args"]:
                                docker_args = action["docker"]["args"]
                            container = self.docker_client.containers.run(
                                image = "%s/%s:%s" % (action["docker_registry"], action["docker"]["name"], action["docker"]["tag"]),
                                detach = True,
                                volumes = {
                                    app_path: {'bind': '/opt/app', 'mode': 'ro'},
                                    workspace: {'bind': '/opt/workspace', 'mode': 'rw'},
                                },
                                command = cmd,
                                **docker_args,
                            )
                            LOG.debug("docker cmd: %s", cmd)
                            stdout_file_path = os.path.join(workspace, "stdout.data")
                            stdout_file = open(stdout_file_path, "w")
                            action["stdout_file_path"] = stdout_file_path
                            action["stdout_file"] = stdout_file
                            action["process"] = container
                            action["start_at"] = datetime.datetime.now()
                        else:
                            if "env" in action:
                                venv_path = os.path.join(action["env"], "bin", "activate")
                                cmd = "cd %s && source '%s' && exec %s '%s' --nodaemon" % (app_path, venv_path, main_path, workspace)
                            else:
                                cmd = "cd %s && exec %s '%s' --nodaemon" % (app_path, main_path, workspace)
                            LOG.debug("cmd: %s", cmd)
                            stdout_file_path = os.path.join(workspace, "stdout.data")
                            stdout_file = open(stdout_file_path, "w")
                            action["stdout_file_path"] = stdout_file_path
                            action["stdout_file"] = stdout_file
                            action["process"] = subprocess.Popen(cmd, shell = True, executable = '/bin/bash', bufsize = -1, stdout = stdout_file, stderr = subprocess.STDOUT)
                            action["start_at"] = datetime.datetime.now()
                    elif isinstance(app_ready, dict): # stop download app failed action
                        action["process"] = None
                        action["start_at"] = datetime.datetime.now()
                        action["result"] = app_ready
                    else: # stop app not ready action
                        if "signal" in action:
                            action["process"] = None
                            action["start_at"] = datetime.datetime.now()
                    self.push_action(action)
                # action already running
                else:
                    action_stage = Stage.running
                    action_status = Status.success
                    action_result = {}
                    now = datetime.datetime.now()
                    end_at = None
                    returncode = None
                    if action["process"] is None: # stop not running action, not need to push back
                        action["update_at"] = now
                        if "signal" in action:
                            if action["signal"] == Signal.cancel:
                                action["canceled"] = True
                            del action["signal"]
                        if "result" in action:
                            action_result = action["result"]
                            del action["result"]
                        action_stage = Stage.finished
                        end_at = now
                        returncode = 0
                        action_status = Status.fail
                        LOG.info("stop action task_id: %s, app_id: %s, name: %s finished: %s", task_id, app_id, name, returncode)
                    elif isinstance(action["process"], docker.models.containers.Container): # run with docker
                        action["process"].reload()
                        if action["process"].status != "exited": # running action, need to push back
                            action["update_at"] = now
                            if "signal" in action:
                                if action["signal"] == Signal.terminate:
                                    action["process"].kill(-Signal.terminate)
                                elif action["signal"] == Signal.kill:
                                    action["process"].kill(-Signal.kill)
                                elif action["signal"] == Signal.cancel:
                                    action["canceled"] = True
                                    action["process"].kill(-Signal.kill)
                                else:
                                    LOG.warning("unknown signal: %s", action["signal"])
                                del action["signal"]
                            LOG.debug("action task_id: %s, app_id: %s, name: %s still running", task_id, app_id, name)
                            self.push_action(action)
                        else: # finished action, not need to push back
                            action_stage = Stage.finished
                            end_at = now
                            r = action["process"].wait()
                            action["stdout_file"].write(json.dumps(r, indent = 4))
                            returncode = r["StatusCode"]
                            output_data_path = os.path.join(workspace, "output.data")
                            if returncode == 0:
                                if os.path.exists(output_data_path) and os.path.isfile(output_data_path):
                                    fp = open(output_data_path, "r")
                                    action_result = json.loads(fp.read())
                                    fp.close()
                            else:
                                action_status = Status.fail
                            if not action["stdout_file"].closed:
                                action["stdout_file"].close()
                            action["process"].remove()
                            LOG.info("action task_id: %s, app_id: %s, name: %s finished: %s", task_id, app_id, name, returncode)
                    else:
                        if action["process"].poll() is None: # running action, need to push back
                            action["update_at"] = now
                            if "signal" in action:
                                if action["signal"] == Signal.terminate:
                                    action["process"].terminate()
                                elif action["signal"] == Signal.kill:
                                    action["process"].kill()
                                elif action["signal"] == Signal.cancel:
                                    action["canceled"] = True
                                    action["process"].kill()
                                else:
                                    LOG.warning("unknown signal: %s", action["signal"])
                                del action["signal"]
                            LOG.debug("action task_id: %s, app_id: %s, name: %s still running", task_id, app_id, name)
                            self.push_action(action)
                        else: # finished action, not need to push back
                            action_stage = Stage.finished
                            end_at = now
                            returncode = action["process"].poll()
                            output_data_path = os.path.join(workspace, "output.data")
                            if returncode == 0:
                                if os.path.exists(output_data_path) and os.path.isfile(output_data_path):
                                    fp = open(output_data_path, "r")
                                    action_result = json.loads(fp.read())
                                    fp.close()
                            else:
                                action_status = Status.fail
                            if not action["stdout_file"].closed:
                                action["stdout_file"].close()
                            LOG.info("action task_id: %s, app_id: %s, name: %s finished: %s", task_id, app_id, name, returncode)

                    if "canceled" in action and action["canceled"]: # canceled action, not need to update status to manager
                        if action_stage == Stage.finished:
                            self.actions_counter -= 1
                            LOG.debug("remove canceled action: %s, actions_counter: %s", action, self.actions_counter)
                    else: # update action status to manager
                        url = "http://%s:%s/action/update" % (CONFIG["manager_http_host"], CONFIG["manager_http_port"])
                        data = {
                            "name": name,
                            "task_id": task_id,
                            "result": action_result,
                            "stage": action_stage,
                            "status": action_status,
                            "node_id": Registrant.instance().config.get("node_id"),
                            "start_at": str(action["start_at"]),
                            "end_at": str(end_at),
                            "returncode": returncode,
                        }
                        LOG.debug("request: %s", url)
                        request = HTTPRequest(url = url, method = "PUT", body = json.dumps(data))
                        r = yield self.async_client.fetch(request)
                        if r.code != 200 or json.loads(r.body.decode("utf-8"))["result"] != Errors.OK: # update status failed
                            if action_stage == Stage.finished:
                                self.push_action(action)
                            LOG.error("update action status failed, task_id: %s, name: %s", task_id, name)
                        else: # update status succeeded
                            if action_stage == Stage.finished:
                                self.actions_counter -= 1
                                LOG.debug("remove action: %s, actions_counter: %s", action, self.actions_counter)
                        LOG.debug("running action: %s", action)
            self.update_heartbeat_data()
            LOG.debug("actions_counter: %s", self.actions_counter)
        except Exception as e:
            LOG.exception(e)

    def close(self):
        try:
            if self.periodic_execute:
                self.periodic_execute.stop()
            LOG.debug("Executor close")
        except Exception as e:
            LOG.exception(e)



# -*- coding: utf-8 -*-

import json
import time
import urllib
import logging

from tornado import web
from tornado import gen

from litepipeline.manager.handlers.base import BaseHandler, BaseSocketHandler
from litepipeline.manager.models.schedules import Schedules
from litepipeline.manager.models.workflows import Workflows
from litepipeline.manager.utils.app_manager import AppManager
from litepipeline.manager.utils.common import file_sha1sum, file_md5sum, Errors, Stage, splitall, JSONLoadError
from litepipeline.manager.config import CONFIG

LOG = logging.getLogger("__name__")


class CreateScheduleHandler(BaseHandler):
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            schedule_name = self.get_json_argument("schedule_name", "")
            source = self.get_json_argument("source", "")
            source_id = self.get_json_argument("source_id", "")
            input_data = self.get_json_argument("input_data", {})
            minute = int(self.get_json_argument("minute", -1))             # [0, 59]
            hour = int(self.get_json_argument("hour", -1))                 # [0, 23]
            day_of_month = int(self.get_json_argument("day_of_month", -1)) # [1, 31]
            day_of_week = int(self.get_json_argument("day_of_week", -1))   # [1, 7] (Sunday = 7)
            enable = True if self.get_json_argument("enable", False) else False
            if (
                    (
                        (source == Schedules.application and AppManager.instance().info(source_id)) or 
                        (source == Schedules.workflow and Workflows.instance().get(source_id))
                    ) and
                    schedule_name and
                    (minute == -1 or (minute >= 0 and minute <= 59)) and
                    (hour == -1 or (hour >= 0 and hour <= 23)) and
                    (day_of_month == -1 or (day_of_month >= 1 and day_of_month <= 31)) and
                    (day_of_week == -1 or (day_of_week >= 1 and day_of_week <= 7))
                ):
                if not isinstance(input_data, dict):
                    raise JSONLoadError("input_data must be dict type")
                schedule_id = Schedules.instance().add(
                    schedule_name,
                    source,
                    source_id,
                    minute = minute,
                    hour = hour,
                    day_of_month = day_of_month,
                    day_of_week = day_of_week,
                    enable = enable,
                    input_data = input_data
                )
                if schedule_id is not False:
                    result["schedule_id"] = schedule_id
                else:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("CreateScheduleHandler, schedule_name: %s, source: %s, source_id: %s", schedule_name, source, source_id)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateScheduleHandler(BaseHandler):
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            schedule_id = self.get_json_argument("schedule_id", "")
            result["schedule_id"] = schedule_id
            data = self.get_json_exists_arguments(
                [
                    "schedule_name",
                    "source",
                    "source_id",
                    "input_data",
                    "minute",        # [0, 59]
                    "hour",          # [0, 23]
                    "day_of_month",  # [1, 31]
                    "day_of_week",   # [1, 7] (Sunday = 7)
                    "enable",
                ]
            )
            if (
                    data and
                    (schedule_id and Schedules.instance().get(schedule_id)) and
                    (("schedule_name" in data and data["schedule_name"] != "") or "schedule_name" not in data) and
                    (
                        "source_id" not in data or
                        ("source_id" in data and data["source"] == Schedules.application and AppManager.instance().info(data["source_id"])) or
                        ("source_id" in data and data["source"] == Schedules.workflow and Workflows.instance().get(data["source_id"]))
                    ) and
                    (("minute" in data and (data["minute"] == -1 or (data["minute"] >= 0 and data["minute"] <= 59))) or "minute" not in data) and
                    (("hour" in data and (data["hour"] == -1 or (data["hour"] >= 0 and data["hour"] <= 23))) or "hour" not in data) and
                    (("day_of_month" in data and (data["day_of_month"] == -1 or (data["day_of_month"] >= 1 and data["day_of_month"] <= 31))) or "day_of_month" not in data) and
                    (("day_of_week" in data and (data["day_of_week"] == -1 or (data["day_of_week"] >= 1 and data["day_of_week"] <= 7))) or "day_of_week" not in data) and
                    (("enable" in data and data["enable"] in [True, False]) or "enable" not in data)
                ):
                if "input_data" in data and not isinstance(data["input_data"], dict):
                    raise JSONLoadError("input_data must be dict type")
                if "app_id" in data:
                    data["application_id"] = data["app_id"]
                    del data["app_id"]
                success = Schedules.instance().update(schedule_id, data)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
            else:
                LOG.warning("invalid arguments")
                Errors.set_result_error("InvalidParameters", result)
            LOG.debug("UpdateScheduleHandler, schedule_id: %s, data: %s", schedule_id, data)
        except JSONLoadError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteScheduleHandler(BaseHandler):
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            schedule_id = self.get_argument("schedule_id", "")
            if schedule_id:
                success = Schedules.instance().delete(schedule_id)
                if not success:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class InfoScheduleHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            schedule_id = self.get_argument("schedule_id", "")
            if schedule_id:
                schedule_info = Schedules.instance().get(schedule_id)
                if schedule_info:
                    result["schedule_info"] = schedule_info
                    if schedule_id in Schedules.instance().cache:
                        result["schedule_cache_info"] = Schedules.instance().cache[schedule_id]
                elif schedule_info is None:
                    Errors.set_result_error("ScheduleNotExists", result)
                else:
                    Errors.set_result_error("OperationFailed", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListScheduleHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            offset = int(self.get_argument("offset", "0"))
            limit = int(self.get_argument("limit", "0"))
            filters = {}
            schedule_id = self.get_argument("schedule_id", "")
            if schedule_id:
                filters["schedule_id"] = schedule_id
            source_id = self.get_argument("source_id", "")
            if source_id:
                filters["source_id"] = source_id
            name = self.get_argument("name", "")
            if name:
                filters["name"] = name
            enable = self.get_argument("enable", "").lower()
            if enable == "true":
                enable = True
            elif enable == "false":
                enable = False
            else:
                enable = ""
            if enable is not "":
                filters["enable"] = enable
            LOG.debug("ListScheduleHandler offset: %s, limit: %s, filters: %s", offset, limit, filters)
            r = Schedules.instance().list(offset = offset, limit = limit, filters = filters)
            result["schedules"] = r["schedules"]
            result["total"] = r["total"]
            result["offset"] = offset
            result["limit"] = limit
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-
'''
LitePipeline: distributed pipeline system
'''

from setuptools import setup

from litepipeline.version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = "litepipeline",
    version = __version__,
    description = "LitePipeline: distributed pipeline system",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/fiefdx/LitePipeline",
    author = "fiefdx",
    author_email = "fiefdx@163.com",
    packages = [
        'litepipeline',
        'litepipeline.tool',
        'litepipeline.manager',
        'litepipeline.manager.db',
        'litepipeline.manager.handlers',
        'litepipeline.manager.models',
        'litepipeline.manager.utils',
        'litepipeline.node',
        'litepipeline.node.handlers',
        'litepipeline.node.utils',
        'litepipeline.viewer',
        'litepipeline.viewer.handlers',
        'litepipeline.viewer.utils',
        'litepipeline.viewer.modules',
    ],
    entry_points = {
        'console_scripts': [
            'litepipeline = litepipeline.tool.litepipeline:main',
            'liteconfig = litepipeline.tool.liteconfig:main',
            'litemanager = litepipeline.manager.manager:main',
            'litenode = litepipeline.node.node:main',
            'liteviewer = litepipeline.viewer.viewer:main',
        ],
    },
    install_requires = [
        "progress >= 1.5",
        "requests >= 2.22.0",
        "tornado",
        "pyYAML",
        "tinydb",
        "sqlalchemy",
        "docker",
        "tornado_discovery",
        "litepipeline_helper >= 0.0.17",
        "litedfs_client >= 0.0.4",
    ],
    include_package_data = True,
    license = "MIT",
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ]
)
